/********** rafael.cpp
		Temporary Rafael shell
				************/

#define	WIN30
#include <owl.h>
#include <filewnd.h>
#include <mdi.h>
#include <listbox.h>
#include <string.h>
#include <stdio.h>
#include <io.h>
#include <math.h>
#include "browwnd.h"
#include "raflist.h"
#include "xserver.h"
#include "rci.h"
#include "rafael.h"

_CLASSDEF(TMDIFileApp)
_CLASSDEF(TMDIFileWindow)
_CLASSDEF(TNFileWindow)
_CLASSDEF(TNBrowser)
_CLASSDEF(TSchedWindow)

// Declare TMDIFileApp, a TApplication descendant
class _CLASSTYPE TMDIFileApp : public TApplication
{
public:
    TMDIFileApp(LPSTR name, HANDLE hInstance,
		  HANDLE hPrevInstance, LPSTR lpCmd,
		  int nCmdShow)
		: TApplication(name, hInstance,
			       hPrevInstance, lpCmd, nCmdShow) {};
    virtual void InitMainWindow();
    virtual void InitInstance();
};

// Declare TNBrowser, a TBrowserWindow descendant
class _CLASSTYPE TNBrowser : public TBrowserWindow
{
public:
  TNBrowser( PTWindowsObject AParent, LPSTR ATitle,
			PTModule AModule = NULL ) :
		TBrowserWindow( AParent, ATitle, AModule ) {};
  virtual void GetWindowClass( WNDCLASS & );
  virtual void ShutDownWindow();
};


// Declare TSchedWindow, a TWindow descendant
class _CLASSTYPE TSchedWindow : public TWindow
{
public:
  SchedListObject	SchedListRoot;
  int			MaxX,MaxY;
  int			RMaxX,RMaxY;
  int			WXSize,WYSize;
  int			SBXOffs,SBYOffs;
  int			XPageSize,YPageSize;
  TSchedWindow( PTWindowsObject AParent, LPSTR ATitle,
		PTModule AModule );
  virtual void Paint( HDC, PAINTSTRUCT _FAR & );
  virtual void ShutDownWindow();
  virtual void GetWindowClass( WNDCLASS & );
  virtual LPSTR GetClassName()
  {
    return "TSchedWindow";
  }
  virtual void WMHScroll( RTMessage Msg ) =
			[ WM_FIRST + WM_HSCROLL ];
  virtual void WMVScroll( RTMessage Msg ) =
			[ WM_FIRST + WM_VSCROLL ];
  virtual void WMSize( RTMessage Msg ) =
			[ WM_FIRST + WM_SIZE ];
};

// Declare TMDIFileWindow, a TMDIFrame descendant
class _CLASSTYPE TMDIFileWindow : public TMDIFrame
{
public:
    WORD ChildNum;
    PTNBrowser		BrowserWindow;
    PTSchedWindow       ScheduleWindow;

    TMDIFileWindow(LPSTR ATitle, LPSTR MenuName);
    virtual void GetWindowClass( WNDCLASS & );
    virtual LPSTR GetClassName()
    {
      return "TMDIFileWindow";
    }
    virtual void EndCompile(RTMessage Msg) =
		 [WM_USER + WM_ENDCOMPILE];
    virtual void MessageLine(RTMessage Msg ) =
		 [WM_USER + WM_COMPMESSAGE];
    virtual void NewFile(RTMessage Msg) =
		 [CM_FIRST + CM_MDIFILENEW];
    virtual void OpenFile(RTMessage Msg) =
		 [CM_FIRST + CM_MDIFILEOPEN];
    virtual void Compile(RTMessage Msg) =
		 [CM_FIRST + CM_COMPILE];
};

// Declare TNFileWindow, TFileWindow descendant
class _CLASSTYPE TNFileWindow : public TFileWindow
{
public:
  TNFileWindow( PTWindowsObject AParent,
		LPSTR ATitle, LPSTR AFileName ) :
		TFileWindow( AParent, ATitle,
			     AFileName ) {};
  virtual void GetWindowClass( WNDCLASS & );
  virtual void WMSetFocus( RTMessage Msg )
		 = [ WM_FIRST + WM_SETFOCUS ];
  virtual BOOL CanClose();
  virtual LPSTR GetClassName()
  {
    return "TNFileWindow";
  }
};

//			Static variables		\\

// Current source
static char SourceFile[MAXPATH];
// Pointer to the actual window
static PTNFileWindow	SourceWindow;
// Full path to the application
static char AppPathName[129];
// Full LISP path name of the application (\s expanded to \\s)
static char LispPathName[129];
// Application instance number
static HANDLE hInst;
// Work variable
static char work[256];
// Main window (copy of TMDIFileApp's MainWindow field
PTWindowsObject	GMainWindow;
// Main menu handle
HMENU MainMenu;
// TRUE if unhandled XL_REQ arrived
BOOL XLREQArrived;
// TRUE if unhandled XL_INCOMPL arrived
BOOL XLINCOMPLArrived;
// Handle of the comm window
HWND	CommWindow;

// TRUE if the compiler is running
BOOL CompilerRunning;
// TRUE if compile is enabled
BOOL CompileEnabled;

char CommClass[] = "RAFAEL.COM";
char Incompl = 0xFF;

//			Static functions		\\

int max(int a, int b)
{
  return( a < b ? b : a );
}

int min(int a, int b)
{
  return( a < b ? a : b );
}

// Gets Xlisp to execute a command. Gets the command
// string (CR-LF terminated), waits for the server
// and returns the reply. Returns NULL if the message
// loop was broken or a pointer to a 0xFF byte if
// the command was incomplete */
LPSTR ExecuteCommand( HWND Window, LPSTR Request )
{
  MSG	msg;

  XDSendRequest( Window, Request );
  XLREQArrived = FALSE;
  XLINCOMPLArrived = FALSE;
  while( !XLREQArrived && !XLINCOMPLArrived &&
	 GetMessage( &msg, NULL, 0 ,0 ) )
  {
    TranslateMessage( &msg );
    DispatchMessage( &msg );
  }

  if( XLREQArrived )
	return XDGetReply( Window );
  else
  if( XLINCOMPLArrived )
	return (LPSTR)&Incompl;
  return NULL;
}

/* Creates the full application file path */
static void MakeAppPathName()
{
  char *  pcFileName;
  int     nFileNameLen;

  nFileNameLen = GetModuleFileName(hInst,AppPathName,128);
  pcFileName = AppPathName + nFileNameLen;

  while (pcFileName > AppPathName )
  {
    if (*pcFileName == '\\' || *pcFileName == ':')
    {
      *(++pcFileName) = '\0';
       break;
    }
    nFileNameLen--;
    pcFileName--;
  }
}

// Expands a pathname to Lisp pathname (expands \s to \\s)
// Gets a pointer to the source and the destination string
void ExpandLispPathName( char *src, char *dest )
{
  while( *src )
  {
    *( dest++ ) = *src;
    if( *src == '\\' )
	*( dest++ ) = '\\';
    ++src;
  }
  *dest = 0;
}

// WndProc of the communication window
long FAR PASCAL _export CommWndProc( HWND hwnd, UINT message,
				     WPARAM wParam, LPARAM lParam )
{
  switch( message )
  {
    case XL_REQ:
	XLREQArrived = TRUE;
	return 0L;

    case XL_INCOMPL:
	XLINCOMPLArrived = TRUE;
	return 0L;

  }
  return DefWindowProc( hwnd, message, wParam, lParam );
}

// Enables compile
void EnableCompile()
{
  EnableMenuItem( MainMenu,CM_COMPILE,MF_BYCOMMAND | MF_ENABLED );
  CompileEnabled = TRUE;
}

// Disables compile
void DisableCompile()
{
  EnableMenuItem( MainMenu,CM_COMPILE,MF_BYCOMMAND | MF_GRAYED );
  CompileEnabled = FALSE;
}

//			TNBrowser			\\

// Redeclaring GetWindowClass to set a new icon
void TNBrowser::GetWindowClass( WNDCLASS &AWndClass )
{
  TBrowserWindow::GetWindowClass( AWndClass );
  AWndClass.hIcon = LoadIcon( hInst,"MSGICON" );
}

// Extending TBrowserWindow's ShutdownWindow to set
// MainWindow's browser field to NULL
void TNBrowser::ShutDownWindow()
{
  ((PTMDIFileWindow)GMainWindow)->BrowserWindow = NULL;
  TBrowserWindow::ShutDownWindow();
}

//			TSchedWindow			\\
TSchedWindow::TSchedWindow( PTWindowsObject AParent,
		LPSTR ATitle, PTModule AModule = NULL ) :
		TWindow( AParent, ATitle, AModule )
{
  Attr.Style |= WS_HSCROLL | WS_VSCROLL;
  RMaxX = 0;
  RMaxY = 0;
  SBXOffs = 0;
  SBYOffs = 0;
  XPageSize = 10;
  YPageSize = 10;
};

// Redeclaring GetWindowClass to set a new icon
void TSchedWindow::GetWindowClass( WNDCLASS &AWndClass )
{
  TWindow::GetWindowClass( AWndClass );
  AWndClass.hIcon = LoadIcon( hInst,"SCHEDICON" );
}

// Horizontal scrollbar response function
void TSchedWindow::WMHScroll( RTMessage Msg )
{
  switch( Msg.WParam )
  {
    case SB_LINEUP:
	--SBXOffs;
	break;

    case SB_LINEDOWN:
	++SBXOffs;
	break;

    case SB_PAGEUP:
	SBXOffs -= XPageSize;
	break;

    case SB_PAGEDOWN:
	SBXOffs += XPageSize;
	break;

   case SB_THUMBPOSITION:
	SBXOffs = Msg.LP.Lo;
	break;
  }
  SBXOffs = max( 0,min( SBXOffs,MaxX ) );
  if( SBXOffs != GetScrollPos( HWindow,SB_HORZ ) )
  {
    SetScrollPos( HWindow , SB_HORZ , SBXOffs , TRUE );
    InvalidateRect( HWindow , NULL , TRUE );
  }
}

// Vertical scrollbar response function
void TSchedWindow::WMVScroll( RTMessage Msg )
{
  switch( Msg.WParam )
  {
    case SB_LINEUP:
	--SBYOffs;
	break;

    case SB_LINEDOWN:
	++SBYOffs;
	break;

    case SB_PAGEUP:
	SBYOffs -= YPageSize;
	break;

    case SB_PAGEDOWN:
	SBYOffs += YPageSize;
	break;

   case SB_THUMBPOSITION:
	SBYOffs = Msg.LP.Lo;
	break;
  }
  SBYOffs = max( 0,min( SBYOffs,MaxY ) );
  if( SBYOffs != GetScrollPos( HWindow,SB_VERT ) )
  {
    SetScrollPos( HWindow , SB_VERT , SBYOffs , TRUE );
    InvalidateRect( HWindow , NULL , TRUE );
  }
}

// Tapping WM_SIZE for client rect size
void TSchedWindow::WMSize( RTMessage Msg )
{
  TWindow::WMSize( Msg );
  WXSize = Msg.LP.Lo;
  WYSize = Msg.LP.Hi;
  SBXOffs = SBYOffs = 0;
  SetScrollPos( HWindow , SB_HORZ , 0 , TRUE );
  SetScrollPos( HWindow , SB_VERT , 0 , TRUE );
}

// Extending TSchedWindow's ShutdownWindow to set
// MainWindow's scheduler window field to NULL
void TSchedWindow::ShutDownWindow()
{
  ((PTMDIFileWindow)GMainWindow)->ScheduleWindow = NULL;
  TWindow::ShutDownWindow();
  SchedListRoot.DeleteSubList();
}

// Paint function
void TSchedWindow::Paint( HDC PaintDC, PAINTSTRUCT _FAR & PaintInfo )
{
  char			text[20];
  PtSchedListObject	p;
  PtSchedListObject	p1;
  HPEN			Pen,OldPen;
  POINT			rect[5];
  int			xc,yc,ycord,xcord,t,t1;
  int			ycord2,xcord2,t2;
  DWORD			fm;
  char			c = '0';
  unsigned		FontWidth;
  float			fac,nx,ny,vx,vy,len;

  fac = 1/sqrt(2);
  Pen = CreatePen( PS_SOLID,1,GetSysColor( COLOR_WINDOWTEXT ) );
  OldPen = SelectObject( PaintDC, Pen );
  SelectObject( PaintDC , GetStockObject( SYSTEM_FONT ) );
  SetBkColor( PaintDC , GetSysColor( COLOR_WINDOW ) );
  SetTextColor( PaintDC , GetSysColor( COLOR_WINDOWTEXT ) );
  fm = GetTextExtent( PaintDC , (LPSTR)&c , 1 );
  FontWidth = LOWORD( fm );
  p = SchedListRoot.ListHead;
  RMaxX = MaxX = 0;
  RMaxY = MaxY = 0;
  ycord = 20-SBYOffs;
  while( p != NULL )
  {
    sprintf( text,"P%u",p->Node );
    TextOut( PaintDC, -SBXOffs+8,ycord+16,
	     text,strlen( text ) );
    p1 = p->ListHead;
    while( p1 != NULL )
    {
      if( p1->Data2 != 0 )
      {
	xcord = 40-SBXOffs+(5*(unsigned)(p1->Data1));
	t = 5*(unsigned)(p1->Data2);
	rect[0].x = xcord;
	rect[0].y = ycord;
	rect[1].x = xcord + t;
	rect[1].y = ycord;
	rect[2].x = xcord + t;
	rect[2].y = ycord + 40;
	rect[3].x = xcord;
	rect[3].y = ycord + 40;
	rect[4].x = xcord;
	rect[4].y = ycord;
	Polyline( PaintDC, rect, 5 );
	t1 = (t-FontWidth) / 2;
	if( p1->Node != 0 )
	{
	  sprintf( text,"%u",p1->Node );
	  TextOut( PaintDC, xcord+t1,ycord+16,
		 text,strlen( text ) );
	}
	if( p1->Node2 != 0 )
	{
	  xcord2 = 40-SBXOffs+(5*(unsigned)(p1->Data3));
	  t2 = (5*(unsigned)(p1->Data2))/2;
	  xcord2 += t2;
	  ycord2 = 40*(p1->Node2)-SBYOffs;
	  xc = xcord+(t/2);
	  yc = ycord+20;
	  MoveTo( PaintDC,xc,yc );
	  LineTo( PaintDC,xcord2,ycord2 );
	  vx = (float)xc - (float)xcord2;
	  vy = (float)yc - (float)ycord2;
	  len = sqrt( vx*vx + vy*vy );
	  vx = vx / len;
	  vy = vy / len;
	  nx = 10*( fac*vx - fac*vy );
	  ny = 10*( fac*vx + fac*vy );
	  LineTo( PaintDC,xcord2+(int)nx,ycord2+(int)ny );
	  MoveTo( PaintDC,xcord2,ycord2 );
	  nx = 10*( fac*vx + fac*vy );
	  ny = 10*( -fac*vx + fac*vy );
	  LineTo( PaintDC,xcord2+(int)nx,ycord2+(int)ny );
	}
	if( ( rect[2].x+SBXOffs ) > MaxX )
		MaxX = rect[2].x + SBXOffs;
      }
      p1 = p1->ForwLink;
    }
    p = p->ForwLink;
    ycord += 40;
  }
  RMaxY = MaxY;
  MaxY = max( 0,ycord+8+SBYOffs-WYSize);
  RMaxX = MaxX;
  MaxX = max( 0,MaxX+8-WXSize);
  SetScrollRange( HWindow,SB_VERT,0,MaxY,TRUE );
  SetScrollRange( HWindow,SB_HORZ,0,MaxX,TRUE );
  SelectObject( PaintDC, OldPen );
  DeleteObject( Pen );
}

//              	TNFileWindow			\\

// Redeclaring GetWindowClass to set a new icon
void TNFileWindow::GetWindowClass( WNDCLASS &AWndClass )
{
  TFileWindow::GetWindowClass( AWndClass );
  AWndClass.hIcon = LoadIcon( hInst,"DOCUICON" );
}

// Redeclaring WM_SETFOCUS handler to save the actual
// file name
void TNFileWindow::WMSetFocus( RTMessage Msg )
{
  TFileWindow::WMSetFocus( Msg );
  if( IsNewFile )
  {
    DisableCompile();
    strcpy( SourceFile,"" );
    SourceWindow = NULL;
  }
  else
  {
    if( !CompilerRunning )
	EnableCompile();
    _fstrncpy( SourceFile,FileName,MAXPATH-1 );
    SourceWindow = this;
  }
}

// Declaring CanClose to disable the compile
// option when the filewindow is closed
BOOL TNFileWindow::CanClose()
{
  BOOL retval;

  if( retval = TFileWindow::CanClose() )
	DisableCompile();
  return retval;
}


//			TMDIFileWindow			\\

// Constructor
TMDIFileWindow::TMDIFileWindow(LPSTR ATitle, LPSTR MenuName)
		   : TMDIFrame(ATitle, MenuName)
{
  ChildNum = 1;
  BrowserWindow = NULL;
  ScheduleWindow = NULL;
}

// Redeclaring GetWindowClass to set a new icon
void TMDIFileWindow::GetWindowClass( WNDCLASS &AWndClass )
{
  TMDIFrame::GetWindowClass( AWndClass );
  AWndClass.hIcon = LoadIcon( hInst,"PICON" );
}

/* Respond to "New" command by constructing, creating, and setting up a
  new TFileWindow MDI child */
void TMDIFileWindow::NewFile(RTMessage)
{
  char ChildName[14];

  wsprintf(ChildName,"");
  GetApplication()->MakeWindow(new TNFileWindow(this, ChildName, ""));
}

/* Respond to "Open" command by constructing, creating, and setting up a
  new TFileWindow MDI child */
void TMDIFileWindow::OpenFile(RTMessage)
{
  char FileName[MAXPATH];
  char ChildName[14];

  wsprintf(ChildName,"" );
  if ( GetApplication()->ExecDialog(new TFileDialog(this, SD_FILEOPEN,
			     _fstrcpy(FileName, "*.sfg"))) == IDOK )
	GetApplication()->MakeWindow(new TNFileWindow(this, ChildName, FileName));
}

// Respond to "Compile" command by starting the compiler
void TMDIFileWindow::Compile( RTMessage Msg )
{
  char	cmdline[MAXPATH+20];

  if( !CompileEnabled )		// May come from accelerator
		return;
  if( SourceWindow != NULL )
	SourceWindow->Save();
  if( *SourceFile )	// If contains name
  {
    if( BrowserWindow == NULL )
    {
      BrowserWindow = new TNBrowser( this,"Messages" );
      BrowserWindow->Attr.X = 8;
      BrowserWindow->Attr.Y = Attr.H - 148;
      BrowserWindow->Attr.H = 96;
      BrowserWindow->Attr.W = Attr.W - 22;
      BrowserWindow->EnableAutoCreate();
      if( GetApplication()->MakeWindow(BrowserWindow) == NULL )
		MessageBox(GMainWindow->HWindow,"Error","Rafael",MB_OK );
    }
    else
	BrowserWindow->Browser->ClearList();
    BringWindowToTop( BrowserWindow->HWindow );
    wsprintf( cmdline, "rcomp %s" , (LPSTR)SourceFile );
    if( WinExec( cmdline , SW_SHOW ) < 32 )
		MessageBox( HWindow, "Compiler exec error",
			    "Rafael" , MB_ICONSTOP | MB_OK );
    else
    {
      DisableCompile();
      CompilerRunning = TRUE;
    }
  }
}

// Compiler end run message
void TMDIFileWindow::EndCompile( RTMessage Msg )
{
  FILE			*sfile;
  unsigned		proc,proc2,node;
  unsigned long		stime,exectime,stime2,exectime2;
  PtSchedListObject	p;
  PtSchedListObject	p1;
  unsigned long		items = 0L;

  EnableCompile();
  CompilerRunning = FALSE;
  if( ScheduleWindow != NULL )
	ScheduleWindow->ShutDownWindow();

  if( ( sfile = fopen( "rcomp.scd", "r" ) ) == NULL )
		return;
  else
  {
    ScheduleWindow = new TSchedWindow( this,"Schedule" );
    ScheduleWindow->EnableAutoCreate();
    if( GetApplication()->MakeWindow(ScheduleWindow) == NULL )
		MessageBox(GMainWindow->HWindow,"Error","Rafael",MB_OK );
    do
      if( fgets( work,255,sfile ) == NULL )
		break;
      else
      if( sscanf( work,"P%u:N%u,%lu,%lu;P%u,%lu,%lu",
		  &proc,&node,&stime,&exectime,
		  &proc2,&stime2,&exectime2 ) < 7 )
		break;
      else
      {
	if( ( p = ScheduleWindow->SchedListRoot.Search( proc ) ) == NULL )
	{
	  p = new SchedListObject( proc,0L,0L,0,0L,0L );
	  ScheduleWindow->SchedListRoot.Insert( p );
	}
	p1 = new SchedListObject( node,stime,exectime,
				proc2,stime2,exectime2 );
	p->Insert( p1 );
	++items;
      }
    while( !feof( sfile ) );
  }
  fclose( sfile );
}

// Message line from the compiler
void TMDIFileWindow::MessageLine( RTMessage Msg )
{
  _fstrcpy( work,RCIGetMessageLine() );
  if( BrowserWindow != NULL )
	BrowserWindow->AddString( work );
}

//			TMDIFileApp			\\

/* Construct the TMDIFileApp's MainWindow of type TMDIFileWindow,
  loading its menu */
void TMDIFileApp::InitMainWindow()
{
  GMainWindow = MainWindow =
		new TMDIFileWindow("Rafael", "Commands");
  ((PTMDIFileWindow)MainWindow)->ChildMenuPos = 4;
}

/* Initialize each MS-Windows application instance, loading an
   accelerator table */
void TMDIFileApp::InitInstance()
{
  WNDCLASS	wndclass;

  TApplication::InitInstance();
  if ( Status == 0 )
  {
	HAccTable = LoadAccelerators(hInstance, "FileCommands");
    if ( HAccTable == 0 )
      Status = EM_INVALIDWINDOW;
  }

  MainMenu = GetMenu( MainWindow->HWindow );
  DisableCompile();
  strcpy( SourceFile,"" );
  SourceWindow = NULL;

// Global variable initialization
  CompilerRunning = FALSE;
  CompileEnabled = FALSE;

// Sets the main window
  RCISetMainWindow( MainWindow->HWindow );
// Creating communication window
  wndclass.style = 0;
  wndclass.lpfnWndProc = CommWndProc;
  wndclass.cbClsExtra = 0;
  wndclass.cbWndExtra = 0;
  wndclass.hInstance = hInst;
  wndclass.hIcon = NULL;
  wndclass.hCursor = NULL;
  wndclass.hbrBackground = NULL;
  wndclass.lpszMenuName = NULL;
  wndclass.lpszClassName = CommClass;
  RegisterClass( &wndclass );

  CommWindow = CreateWindow( CommClass,NULL,WS_CHILD,
			       0,0,0,0,MainWindow->HWindow,NULL,
			       hInst,NULL );

// Starts the server and loads the database
  if( XDStartServer( CommWindow ) )
  {
    MessageBox( NULL,"XLisp server cannot be launched",
		    "Rafael",MB_ICONSTOP | MB_OK );
    Status = 1;
  }
  else
  {
    MakeAppPathName();
    ExpandLispPathName( AppPathName,LispPathName );
    strcpy( work,"(load '" );
    strcat( work, LispPathName );
    strcat( work, "rafdb.lsp" );
    strcat( work, ")\n" );
    if( ExecuteCommand( CommWindow, (LPSTR)work )
				!= NULL )
	XDDeleteReply( CommWindow );
    else
	Status = 1;

    if( !Status )
    {
      strcpy( work,"(load '" );
      strcat( work, LispPathName );
      strcat( work, "rafcomp.lsp" );
      strcat( work, ")\n" );
      if( ExecuteCommand( CommWindow, (LPSTR)work ) != NULL )
	XDDeleteReply( CommWindow );
      else
	Status = 1;
    }

    if( !Status )
    {
      strcpy( work,"(load '" );
      strcat( work, LispPathName );
      strcat( work, "rafhw.lsp" );
      strcat( work, ")\n" );
      if( ExecuteCommand( CommWindow, (LPSTR)work ) != NULL )
	XDDeleteReply( CommWindow );
      else
	Status = 1;
    }
  }
}

// Run the MDIFileApp
int PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance,
		   LPSTR lpCmd, int nCmdShow)
{
  TMDIFileApp MDIFileApp("MDIFileApp", hInstance, hPrevInstance,
		lpCmd, nCmdShow);

  hInst = hInstance;
  MDIFileApp.Run();
  XDTerminateServer();
  DestroyWindow( CommWindow );
  return MDIFileApp.Status;
}
