; Rafael experimental database
;   Contains only TMS320C30 operators
;   Written by Gabor Paller

; Banner function. Gets the device name and writes the
; program banner into outtext
(defun banner(device)
    (setq outtext "")
    (if (equal device 'c30)
         (setq outtext (concatenate 'STRING 
                                ";%s\n"
                                ";\t\tDevice type : TMS320C30\n"
                                ";Rafael code generator, Version 1.1\n"
                                ";Task requires %lu words of memory and takes %lu cycles to execute\n"
                                "\n\n\n"
          )
        )
    )
    outtext
)

; Library initialziation function. The compiler calls this
; function to get the library initialize its variables
(defun init-lib()
        (setq local-labels 0)
)

; Returns the equ statement for the given device. 
; Gets the device ID, returns the equ string
(defun get-equ(device)
    (setq outtext "")
    (if (equal device 'c30)
	(setq outtext "VAR%lu\t.set\tarena+%s")
    )
    outtext
)
        
; Returns the remark string for the given device. 
; Gets the device ID, returns the equ string 
(defun get-rem(device)
    (setq outtext "")
    (if (equal device 'c30)
        (setq outtext "; ")
    )
    outtext
)

; Returns the header string for the constructor section
; Gets the device ID, returns the text
(defun get-begin-constr(device)
    (setq outtext "")
    (if (equal device 'c30)
        (setq outtext "; Constructor section\n\t.def\tconstr\nconstr:" )
    )
    outtext
)

; Returns the footer string for the constructor section
; Gets the device ID, returns the text
(defun get-end-constr(device)
    (setq outtext "")
    (if (equal device 'c30)
        (setq outtext "\tretsu\n" )
    )
    outtext
)

; Returns the header string for the operator section
; Gets the device ID, returns the text
(defun get-begin-op(device)
    (setq outtext "")
    (if (equal device 'c30)
        (setq outtext "; Operator section\n\t.def\toperators\noperators:\n" )
    )
    outtext
)

; Returns the footer string for the operator section
; Gets the device ID, returns the text
(defun get-end-op(device)
    (setq outtext "")
    (if (equal device 'c30)
        (setq outtext "\tretsu\n" )
    )
    outtext
)

; Returns the header string for the destructor section
; Gets the device ID, returns the text
(defun get-begin-destr(device)
    (setq outtext "")
    (if (equal device 'c30)
        (setq outtext "; Destructor section\n\t.def\tdestr\ndestr:\n" )
    )
    outtext
)

; Returns the header string for the destructor section
; Gets the device ID, returns the text
(defun get-end-destr(device)
    (setq outtext "")
    (if (equal device 'c30)
        (setq outtext "\tretsu\n" )
    )
    outtext
)

; Gets the block head text for the given device.
; Gets the device ID and the block name, returns the text
(defun get-begin-block(device blockname)
    (setq outtext "")
    (if (equal device 'c30)
        (setq outtext (concatenate 'STRING
                        "; Head of block "
                        (string blockname)
                        "\n"
                        (string blockname)
                        ":\n" )
        )
    )
    outtext
)

; Gets the block invocation text for the given device.
; Gets the device ID and the block name, returns the text
(defun get-invoc-block(device blockname cond)
  (let ( ( iftext "" ) (endiftext ""))
    (setq outtext "")
    (if (equal device 'c30)
      (progn
        (if (not (null cond))
            (progn
              (setq iftext (concatenate 'STRING
                                "\tif\t"
                                (string cond)
                                "\n"
                           )
              )
              (setq endiftext "\tendif\n")
            )
        )
        (setq outtext (concatenate 'STRING
                "; Calling block "
                (string blockname)
                "\n"
                iftext
                "\tcall\t"
                (string blockname)
                "\n"
                endiftext
          )
        )
      )
    )
  )
  outtext
)

; Returns the block end text for the given device.
; Gets the device name and the block name to be closed
(defun get-end-block(device blockname)
    (setq outtext "")
    (if (equal device 'c30)
        (setq outtext (concatenate 'STRING
                    "\tretsu\n"
                    "; End of block "
                    (string blockname)
                    "\n"
                    )
        )
    )
    outtext
)
 
; Returns the sender communication procedure text
; Gets the sender device number and device ID, the receiver
; device number and device ID, the address of the  variable
; to be sent and the size of the variable to be sent
; and returns the text of the comm procedure in outtext
(defun send-var(from fid to tid var size)
  (setq outtext (concatenate 'STRING
                    "\tcomm "
                    (string var)
                    " from "
                    (string from)
                    " to "
                    (string to)
                    "\n"
                )
  )
  nil 
)

; Returns the receiver communication procedure text
; Gets the sender device number and device ID, the receiver
; device number and device ID, the address of the variable
; to be received and the size of the variable to be received
; and returns the text of the comm procedure in outtext
(defun rec-var(from fid to tid var size)
  (setq outtext (concatenate 'STRING
                    "\tcomm "
                    (string var)
                    " from "
                    (string from)
                    " to "
                    (string to)
                    "\n"
                )
  )
  nil 
)

; Returns the text for the loop head operator. Gets the
; device ID, the loop ID and the iteration number
(defun get-loop-head(device loopid repeat)
  (setq outtext "")
  (if (equal device 'c30)
    (setq outtext (concatenate 'STRING
                    "\t.data\n"
                    "loopvar"
                    (string loopid)
                    ":\t.word\t0\n"
                    "\t.text\n"
                    "\tldi\t"
                    (string repeat)
                    ",r0\n"
                    "\tsti\tr0,@loopvar"
                    (string loopid)
                    "\nloop"
                    (string loopid)
                    ":\n"
      )
    )
  )
  outtext
)

; Returns the text for the loop end operator. Gets the
; device ID and the loop ID.
; If the return value is an empty string, the compiler
; will simply repeat the operations inside the loop.
(defun get-loop-end(device loopid)
  (setq outtext "")
  (if (equal device 'c30)
    (setq outtext (concatenate 'STRING
                "\n\tldi\t@loopvar"
                (string loopid)
                ",r0\n"
                "\tsubi\t1,r0\n"
                "\tsti\tr0,@loopvar"
                (string loopid)
                "\n\tbnz\tloop"
                (string loopid)
                "\n"
      )
    )
  )
  outtext
)

;;;                                 ;;;
;;;           Operator library      ;;;
;;;                                 ;;;
 
(setq inpprobe '(( (c30 c30iprobe nil) )
              ()
              ( (type1 size c 1 1) )
              ( (c30 size) )
              ( size )
              ( (c30 nil) )
              ( (c30 nil) )
             )
)
 
(setq outprobe '(( (c30 c30oprobe nil) )
              ( (type1 size c 1 1) )
              ( (type1 size c 1 1) )
              ( (c30 size) )
              ( size )
              ( (c30 nil) )
              ( (c30 nil) )
             )
)
 
(setq const '(( (c30 c30const nil) )
              ()
              ( (type1 (length cnst) c 1 1) )
              ( (c30 0) )
              ( cnst )
              ( (c30 nil) )
              ( (c30 nil) )
             )
)

(setq add '(( (c30 c30add nil) )
            ( (type1 n) (type1 n) )
            ( (type1 n a 1 1) )
            ( (c30 (+ (* 2 n) 10) )
              (dsp96k (+ (* 2 n) 5) )
            )
            ( n )
            ( (c30 nil) )
            ( (c30 nil) )
           )
)

(setq mul '(( (c30 c30mul nil) )
            ( (type1 n) (type1 n) )
            ( (type1 n a 1 1) )
            ( (c30 (* 2 n) ) )
            ( n )
            ( (c30 nil) )
            ( (c30 nil) )
           )
)

(setq delay '(( (c30 c30delay nil) )
              ( (type1 size) )
              ( (type1 size a 1 1) )
              ( (c30 size) )
              ( size taps )
              ( (c30 c30delay-constr) )
              ( (c30 nil) )
             )
)

(setq interpolator
            '(( (c30 c30intr nil) )
             ( (type1 size) )
             ( (type1 size a fact 1) )
             ( (c30 size) )
             ( fact size )
             ( (c30 c30intr-constr) )
             ( (c30 nil) )
            )
)

(setq decimator
            '(( (c30 c30dec nil) )
             ( (type1 size) )
             ( (type1 size a 1 fact) )
             ( (c30 size) )
             ( fact size )
             ( (c30 nil) )
             ( (c30 nil) )
            )
)

(setq concentrator
            '(( (c30 c30conc nil) )
             ( (type1 size) )
             ( (type1 size a 1 size) )
             ( (c30 1) )
             ( size )
             ( (c30 c30conc-constr) )
             ( (c30 nil) )
            )
)

(setq deconcentrator
            '(( (c30 c30deconc nil) )
             ( (type1 size) )
             ( (type1 size a size 1) )
             ( (c30 1) )
             ( size )
             ( (c30 c30conc-constr) )
             ( (c30 nil) )
            )
)

; C30 strategy routines section
; input probe strategy
(defun c30iprobe( par input output nodenum )
  (setq outtext (concatenate 'STRING
            "\tldi\t"
            (string (first output))
            ",ar0\n"
            "\tcall\tprobe_"
            (string nodenum)
            "\n"
    )
  )
)

; C30 output probe strategy
(defun c30oprobe( par input output nodenum )
  (setq outtext (concatenate 'STRING
            "\tldi\t"
            (string (first input))
            ",ar0\n"
            "\tcall\tprobe_"
            (string nodenum)
            "\n"
    )
  )
)

; C30 const strategy
(defun c30const( par inputs output nodenum )
  (setq outtext (concatenate 'STRING
                    "\t.data\n"
                    (string (car output))
                    ":\n")
  )
  (dolist (n (car par))
    (setq outtext (concatenate 'STRING
                      outtext
		      "\t.float\t"
                      (string n)
                      "\n")
    )
  )
  (setq outtext (concatenate 'STRING
                      outtext
                      "\n\t.text\n"
                )
  )
  nil
)

; C30 Add strategy
(defun c30add(par inputs output nodenum)
    (if (eql (first par) 1)
            (setq outtext (concatenate 'STRING
                                 "\tldf\t"
                                 (string (first inputs))
                                 ",r0\n"
                                 "\tldf\t"
                                 (string (second inputs))
                                 ",r1\n"
                                 "\taddf3\tr0,r1,r1\n"
                                 "\tstf\tr1,"
                                 (string (first output))
			)
          )
          (progn 
            (setq outtext (concatenate 'STRING
                                 "\t.data\n"
                                 "ptra"
                                 (string nodenum)
                                 ":\t.word\t"
                                 (string (first inputs))
                                 "\nptrb"
                                 (string nodenum)
                                 ":\t.word\t"
                                 (string (second inputs))
                                 "\nptro"
                                 (string nodenum)
                                 ":\t.word\t"
                                 (string (first output))
                                 "\n\t.text\n\tldi\t@ptra"
                                 (string nodenum)
                                 ",ar0\n"
                                 "\tldi\t@ptrb"
                                 (string nodenum)
                                 ",ar1\n"
                                 "\tldi\t@ptro"
                                 (string nodenum)
                                 ",ar2\n"
				 "\tldi\t"
                                 (string (- (first par) 1))
                                 ",rc\n" 
                                 "\trptb\tll"
				 (string local-labels)
				 "\n"
				 "\taddf3\t*ar0++(1),*ar1++(1),r1\n"
				 "ll"
				 (string local-labels)
				 ":"
				 "\tstf\tr1,*ar2++(1)\n"
			)
	    )
	    (setq local-labels (+ local-labels 1))
	  )
    )
)

; C30 Multiply strategy
(defun c30mul(par inputs output nodenum)
    (if (eql (first par) 1)
          (setq outtext (concatenate 'STRING
                                 "\tldf\t"
                                 (string (first inputs))
				 ",r0\n"
				 "\tldf\t"
				 (string (second inputs))
                                 ",r1\n"
                                 "\tmpyf3\tr0,r1,r1\n"
                                 "\tstf\tr1,"
                                 (string (first output))
                        )
          )
          (progn 
	    (setq outtext (concatenate 'STRING
                                 "\t.data\n"
                                 "ptra"
                                 (string nodenum)
                                 ":\t.word\t"
                                 (string (first inputs))
                                 "\nptrb"
                                 (string nodenum)
                                 ":\t.word\t"
                                 (string (second inputs))
                                 "\nptro"
                                 (string nodenum)
                                 ":\t.word\t"
                                 (string (first output))
                                 "\n\t.text\n\tldi\t@ptra"
                                 (string nodenum)
                                 ",ar0\n"
                                 "\tldi\t@ptrb"
                                 (string nodenum)
                                 ",ar1\n"
                                 "\tldi\t@ptro"
                                 (string nodenum)
                                 ",ar2\n"
				 "\tldi\t"
				 (string (- (first par) 1))
                                 ",rc\n" 
                                 "\trptb\tll"
                                 (string local-labels)
                                 "\n" 
                                 "\tmpyf3\t*ar0++(1),*ar1++(1),r1\n"
                                 "ll"
                                 (string local-labels)
                                 ":" 
				 "\tstf\tr1,*ar2++(1)\n"
			)
            )
            (setq local-labels (+ local-labels 1))
          )
    ) 
)

; C30 Delay strategy
(defun c30delay(par input output nodenum)
    (if (not (null output))     ;If get-delay
	(progn
	  (setq outtext (concatenate 'STRING
                    "\t.data\nptro"
                    (string nodenum)
                    ":\t.word\t"
                    (string (first output))
                    "\n\t.text\n"
                    "\tldi\t@dptr"
		    (string nodenum)
		    ",ar0\n"
		    "\tldi\t@dsize"
		    (string nodenum)
		    ",bk\n"
		    "\tldi\t@ptro"
		    (string nodenum)
		    ",ar1\n"
	    )
	  )
	  (if (eql (car par) 1)
	    (setq outtext (concatenate 'STRING
		    outtext
		    "\tldf\t*ar0,r0\n"
		    "\tstf\tr0,*ar1\n"
	      )
	    )
	    (progn
	      (setq outtext (concatenate 'STRING  ;If vector
		    outtext
		    "\tldi\t@dsize"
		    (string nodenum)
		    ",lc\n"
		    "\trptb\tll"
		    (string local-labels)
		    "\n\tldf\t*ar0++(1),r0\n"
		    "ll"
		    (string local-labels)
		    ":\tstf\tr0,*ar1++(1)\n" )
	      )         ;setq
	      (setq local-labels (+ local-labels 1))
	    )
	  )             ;if (eql) ...
	)               ;progn
	(progn              ;if put-delay
	  (setq outtext (concatenate 'STRING
                    "\t.data\nptri"
                    (string nodenum)
                    ":\t.word\t"
                    (string (first input))
                    "\n\t.text\n"
		    "\tldi\t@dptr"
		    (string nodenum)
		    ",ar0\n"
		    "\tldi\t@dsize"
		    (string nodenum)
		    ",bk\n"
		    "\tldi\t@ptri"
		    (string nodenum)
		    ",ar1\n"
	    )
	  )
	  (if (eql (car par) 1)
	    (setq outtext (concatenate 'STRING
		    outtext
		    "\tldf\t*ar1++(1),r0\n"
		    "\tstf\tr0,*ar0\n"
		    "\tstf\tar1,@dptr"
		    (string nodenum)
		    "\n"
	      )
	    )
	    (progn
	      (setq outtext (concatenate 'STRING  ;If vector
		    outtext
		    "\tldi\t@dsize"
		    (string nodenum)
		    ",lc\n"
		    "\trptb\tll"
		    (string local-labels)
		    "\n\tldf\t*ar1++(1),r0\n"
		    "ll"
		    (string local-labels)
		    ":\tstf\tr0,*ar0++(1)\n"
		    "\tstf\tar1,@dptr"
		    (string nodenum)
		    "\n"
		)
	      )         ;setq
	      (setq local-labels (+ local-labels 1))
	    )           ;progn
	  )             ;if (eql) ...
	)               ;progv ;put-delay
    )                   ;if (null output)
)

; C30 Delay constructor
(defun c30delay-constr(par input output nodenum)
    (if (not (null output))     ;If get-delay
        (progn
          (setq outtext (concatenate 'STRING
		    "\n\t.data\n\t\n"
		    "dptr"
		    (string nodenum)
		    ":\t.word\tdelay"
		    (string nodenum)
		    "\ndsize"
		    (string nodenum)
		    ":\t.word\t"
		    (string (* (first par) (second par)))
		    "\n\t.bss\tdelay"
		    (string nodenum)
                    ","
		    (string (* (first par) (second par)))
		    "\n\t.text\n"
                    "\tldi\t@dptr"
                    (string nodenum)
                    ",ar0\n"
                    "\tldi\t@dsize"
                    ",lc\n"
                    "\tldf\t0,r0\n"
                    "\trptb\tll"
                    (string local-labels)
                    "\nll"
                    (string local-labels)
                    ":\tstf\tr0,*ar0++(1)\n"
            )
          )     ;setq
          (setq local-labels (+ local-labels 1))
          1                     ;"Text generated" flag
        )       ;progn
        nil     ;if put-delay, returns nil        
    )                           ;if(not (null ...
)

; C30 Interpolator strategy
(defun c30intr(par input output nodenum)
        (setq outtext (concatenate 'STRING
                        "\tldi\t@ptri"
                        (string nodenum)
                        ",ar0\n"
                        "\tldi\t@ptro"
                        (string nodenum)
                        ",ar1\n"
			"\tldi\t"
                        (string (- (second par) 1))
                        ",rc\n" 
                        "\tldi\t@intr"
                        (string nodenum)
                        ",r0\n\taddi\t1,r0\n"
                        "\tsti\tr0,@intr"
                        (string nodenum)
                        "\n\tsubi\t"
                        (string (- (first par) 1))
                        ",r0\n\tbz\tll"
                        (string local-labels)
                        "\n\trep\t"
                        (string (second par))
			"\n\tstf\t0,*ar1++(1)\n"
                        "\tb\tll"
                        (string (+ 1 local-labels))
                        "\nll"
                        (string local-labels)        
                        ":\trptb\tll"
			(string (+ 2 local-labels))
			"\n"
			"\tldf\t*ar0++(1),r1\n"
			"ll"
			(string (+ 2 local-labels))
			":"
			"\tstf\tr1,*ar1++(1)\nll"
                        (string (+ 1 local-labels))
                        ":\n"
                     ) ;concatenate
        )
        (setq local-labels (+ local-labels 3))
)

; C30 Interpolator constructor
(defun c30intr-constr(par input output nodenum)
        (setq outtext (concatenate 'STRING
                "\t.data\nintr"
                (string nodenum)
                ":\t.word\t0\n"
                "ptri"
                (string nodenum)
                ":\t.word\t"
                (string (first input))
                "\n\tptro"
                (string nodenum)
                "\t.word\t"
                (string (first output))
                "\n\t.text\n"
                )
        )
        1       ;"Text generated" flag
)

; C30 Decimator strategy
(defun c30dec(par input output nodenum)
        (setq outtext (concatenate 'STRING
                        "\t.data\nptri"
                        (string nodenum)
                        ":\t.word\t"
                        (string (first input))
                        "\n\tptro"
                        (string nodenum)
                        "\t.word\t"
                        (string (first output))
                        "\n\t.text\n\tldi\t@ptri"
                        (string nodenum)
                        ",ar0\n"
                        "\tldi\t@ptro"
                        (string nodenum)
                        ",ar1\n"
			"\tldi\t"
                        (string (- (second par) 1))
                        ",rc\n" 
                        "\trptb\tll"
			(string local-labels)
			"\n"
			"\tldf\t*ar0++(1),r1\n"
			"ll"
			(string local-labels)
			":"
			"\tstf\tr1,*ar1++(1)\n"
                     ) ;concatenate
        )                
        (setq local-labels (+ local-labels 1))
)

; C30 Concentrator/deconcentrator constructor
(defun c30conc-constr(par input output nodenum)
        (setq outtext (concatenate 'STRING
                "\t.data\nconc"
                (string nodenum)
                ":\t.word\t0\n"
                "ptri"
                (string nodenum)
                ":\t.word\t"
                (string (first input))
                "\nptro"
                (string nodenum)
                ":\t.word\t"
                (string (first output))
                "\n\t.text\n"
                )
        )
        1       ;"Text generated" flag
)

; C30 Concentrator strategy 
(defun c30conc(par input output nodenum)
        (setq outtext (concatenate 'STRING
                "\tldi\t@ptri"
                (string nodenum)
                ",ar0\n\tldi\t*ar0,r0\n"
                "\tldi\t@conc"
                (string nodenum)
                ",ar1\n\tldi\t"
                (string (first par))
                ",bk\n\tldi\t@ptro"
                (string nodenum)
                ",ir1\n\tsti\tr0,*+ar1(ir1)\n"
                "\tnop\t*ar1++(1)%\n"
                "\tsti\tar1,@conc"
                (string nodenum)
                "\n"
                )
        )
)

; C30 Deconcentrator strategy 
(defun c30deconc(par input output nodenum)
        (setq outtext (concatenate 'STRING
                "\tldi\t@ptri"
                (string nodenum)
                ",ar0\n\tldi\t@conc"
                (string nodenum)
                ",ar1\n\tldi\t"
                (string (first par))
                ",bk\n\tldi\t@ptro"
                (string nodenum)
                ",ir1\n\tldi\t*+ar1(ir1),r0\n"
                "\tnop\t*ar1++(1)%\n"
                "\tsti\tar1,@conc"
                (string nodenum)
                "\n\tsti\tr0,*ar0\n"
                )
        )
)

;                       END HEADER - RAFDB.LSP
