; Rafael compiler, hardware descriptor part
; This part depends totally on the target hardware.
; This file describes an imaginary system with two TMS320C30 DSPs
; *** This file must be created by the user who describes the hardware ***
 

(setq target '((1 c30 16000)
	       (2 c30 16000)
	       (3 c30 16000))
)

; Communication cost routine. Calling :
; (get-commcost <frproc> <toproc> <op> <size>) where frproc is the
; number of the sender, toproc is the destination proc,
; op is R or S depending whether the operation is executed
; on the sender or receiver side, size is the blocksize
; of the variable sent. 
; The routine must return a three-element list where the
; first element describes the communication time on
; the given path, the second element is the survive
; time of the variable on the sender proc which is sent,
; the third element is 0 or 1, it is 1 if the sender
; and receive activity must be synchronized (scheduled
; at the same time)
(defun get-commcost(frproc toproc op size)
  (if (or (and (eql frproc 1) (eql toproc 2))
          (and (eql frproc 2) (eql toproc 1))
          (and (eql frproc 2) (eql toproc 3))
          (and (eql frproc 3) (eql toproc 2))
      )
    (list (+ size 10) 0 1)
    (if (or (and (eql frproc 1) (eql toproc 3))
          (and (eql frproc 3) (eql toproc 1))
      )
      (list (* (+ size 10) 2) 0 1)
    )
  )
)

; Communication cost estimation routine. Calling :
; (get-commcost-est <frproc> <toproc> <size>) where frproc is the
; number of the sender, toproc is the destination proc,
; size is the blocksize of the variable sent. 
; The routine must return a more or less correct estimation
; of the time of communication of the given blocksize between
; the two processors. It must include the cost of routing
; if necessary.
(defun get-commcost-est(frproc toproc size)
  (if (or (and (eql frproc 1) (eql toproc 2))
          (and (eql frproc 2) (eql toproc 1))
          (and (eql frproc 2) (eql toproc 3))
          (and (eql frproc 3) (eql toproc 2))
      )
    (+ size 10)
    (if (or (and (eql frproc 1) (eql toproc 3))
          (and (eql frproc 3) (eql toproc 1))
      )
      (* (+ size 10) 2)
    )
  )
)

; This routine returns the total number of communication
; channels available in the target system. The routine
; is called at compiler setup time when the compiler
; allocates the channel descriptor blocks for flagging
; the status of the channels
(defun get-channelnum()
 3
)

; This routine returns the number of the channel used
; for communication between a given processor pair.
; Calling form : (get-channel(<frproc> <toproc>))
; where frproc and toproc are the sender and receiver,
; respectively.
(defun get-channel(frproc toproc)
  (if (or (and (eql frproc 1) (eql toproc 2))
          (and (eql frproc 2) (eql toproc 1))
      )
    1
    (if (or (and (eql frproc 2) (eql toproc 3))
          (and (eql frproc 3) (eql toproc 2))
        )
      2
      (if (or (and (eql frproc 1) (eql toproc 3))
              (and (eql frproc 3) (eql toproc 1))
          )
        3
      )
    )
  )
)

