/******** rc1.cpp
		Rafael compiler,part 1
		Written by Gabor Paller
				******/

#include	<windows.h>
#include	<stdio.h>
#include	<ctype.h>
#include	<string.h>
#include	"robject.h"
#include	"rcomp.h"
#include	"xserver.h"
#include	"rci.h"

FILE	*DFile;

// Pointer to the actual block descriptor
PtBlock		ActBlock;

// Root block
Block		RootBlock( "ROOT",TRUE );

// Processor list. Stores the available physical
// devices on the first level and the communication
// costs on the second level
Processor	ProcessorList( 0,0L,"" );

// Number of channels available for communication
unsigned	NumberOfChannels;

INPUTTYPE	MaxProbeNum;

// Limit upscale and downscale values set by CheckIPF
// and used for finding out which IPF subgraph should
// be scheduled only once
unsigned LUpScale,LDownScale;
float	 LFactor;
// TRUE if dynamic scheduling at root level
BOOL	IsDynamic;

// Message file
extern FILE		*MsgFile;
// Source file
extern FILE		*SrcFile;
// Line counter
extern unsigned long	LineCtr;
// Error counter
extern unsigned ErrCtr;
// Warning counter
extern unsigned WarnCtr;
// Handle of the main window
extern HWND MainWindow;
// Handle of the comm window
extern HWND CommWindow;

// Variable ID counter
unsigned long	VarID;

// Actual line to be interpreted
char	ActLine[129];
// Actual character pointer in the line
char	*ActPtr;

// work string variable
extern char work[1024];
// work string variable 2
char	work1[140];
// work string variable 3
char	work2[1024];
// End-of-file flag
BOOL	eof;
// TRUE if line numbers should be incorporated into error
// messages
BOOL ErrorWithLineNumbers;

// Imported functions from RCOMP.CPP
void FatalError();
void SendMessageLine( char * );
void UpdateDlgText( int,char * );
LPSTR DRCIExecuteCommand( HWND,LPSTR );

// Imported functions from RC2.CPP
void ResetAllNodeFlags();
void ScheduleBlockTree();

// Imported functions form RC3.CPP
void Codegen();

// Updates error counter
void IncError()
{
  char	work[10];

  ++ErrCtr;
  sprintf( work,"%u",ErrCtr );
  UpdateDlgText( MD_ERRTEXT,work );
}

// Writes a line into the message output
void ErrorMessage( char msg[] )
{
  char	work[140];

  IncError();
  if( ErrorWithLineNumbers )
	sprintf( work,"Line %lu : %s",LineCtr,msg );
  else
	sprintf( work,"Error : %s",msg );
  SendMessageLine( work );
}

// Error message without line number
void Error2Message( char msg[] )
{
  IncError();
  sprintf( work,"in block %s : %s",ActBlock->BlockName,
				   msg );
  SendMessageLine( work );
}

// FatalError with parameter
void FatalError( char msg[] )
{
  Error2Message( msg );
  FatalError();
}

// Updates warning counter
void IncWarning()
{
  char	work[10];

  ++WarnCtr;
  sprintf( work,"%u",WarnCtr );
  UpdateDlgText( MD_WARNTEXT,work );
}

// Writes a line into the message output
void WarningMessage( char msg[] )
{
  char	work[140];

  IncWarning();
  sprintf( work,"Warning : %s",msg );
  SendMessageLine( work );
}

// Reads a line into ActLine. Returns TRUE if end of
// file. Sets back ActPtr to the beginning of the line
BOOL ReadLine()
{
  char	*p;

  ++LineCtr;
  ActPtr = ActLine;
  eof = ( fgets( ActLine, 128, SrcFile ) == NULL );
  if( !eof )
  {
    p = ActLine;
    while( *p )
    {
      if( *p == '\n' )
	*p = 0;
      p++;
    }
  }
  return eof;
}

// Skips the spaces from the actual position to
// the first non-white character
void SkipSpc()
{
  while( ( *ActPtr == ' ' ) || ( *ActPtr == 0x09 ) )
	++ActPtr;
}

// Fit routine. Returns TRUE if the parameter string
// can be fitted to the remaining part of the line.
// Pulls the actual character pointer to the end of
// the string fitted.
BOOL fit( char string[] )
{
  char	*p;

  SkipSpc();
  p = ActPtr;
  while( *string )
    if( toupper( *( p++ ) ) != *( string++ ) )
		return FALSE;
  ActPtr = p;
  return TRUE;
}

// Gets a white-character limited string into the
// 'work' variable. Returns a pointer to work. Its
// parameter is the maximum length of the string
char *GetString( int len )
{
  char	*p,*p1;

  SkipSpc();
  p = ActPtr;
  p1 = work;
  while( ( len-- ) && ( *p != ' ' ) && ( *p != 0x09 )
	 && ( *p != 0 ) && ( *p != '\n' ) )
	*( p1++ ) = *( p++ );
  *p1 = 0;
  ActPtr = p;
  return work;
}

// Searches the internal node of a block. Gets the number
// of I/O of the node, the node type and a pointer
// to the block
PtNodeObject BlockIntNode( NODETYPE InputNum,int IType,
				PtBlock bptr )
{
  PtPTrans	pt;
  PtNodeInfo	infoptr;
  PtNodeObject  nodeptr;
  NODETYPE	nn,in;

  pt = (bptr->PTList).Search( InputNum,IType );
  if( pt == NULL )
  {
    wsprintf( work1,"BlockIntNode : Input %u of node %u in block %s not found",
		nodeptr->InputNum,nodeptr->NodeNum,
		(LPSTR)bptr->BlockName );
    FatalError( work1 );
  }
  nn = pt->PNodeNum;
  in = pt->PIONum;
  infoptr = (PtNodeInfo)(bptr->NodeList).Search( nn,OPTYPE_ALL );
  if( infoptr == NULL )
  {
    wsprintf( work1,"BlockIntNode : Node %u not found",
			nn );
    FatalError( work1 );
  }
  nodeptr = (PtNodeObject)(infoptr->GraphPtr);
  if( IType == PTRANS_INPUT )
	nodeptr = nodeptr->Search( in,NOT_VALID );
  else
	nodeptr = nodeptr->Search( NOT_VALID,in );
  if( nodeptr == NULL )
  {
    wsprintf( work1,"BlockIntNode : Input %u of node %u not found",
			  in,nn );
    FatalError( work1 );
  }
  return nodeptr;
}

// Inserts a probe into the NodeInfo list. Evaluates
// the remaining part of the line and creates a node
// entry.
void InsertProbe()
{
  PtNodeInfo	p;
  PtNodeObject	np,np2;
  PtProcessor	pp;
  NODETYPE	node;
  unsigned	in,on,procnum;
  char		ptype[32];
  unsigned	us = 1,ds = 1;

  GetString( 2 );	// Reading I/O flag
  switch( toupper( *work ) )
  {
    case 'I':
	in = 0; on = 1;
	strcpy( ptype,"INPPROBE" );
	break;

    case 'O':
	in = 1; on = 0;
	strcpy( ptype,"OUTPROBE" );
	break;

    default:
	sprintf( work1,"Invalid I/O qualifier : %s",work );
	ErrorMessage( work1 );
	return;
  }

  GetString( 10 );	// Reading node number
  if( sscanf( work,"%u",&node ) < 1 )
  {
    sprintf( work1,"Syntax error : '%s' is not a node number",work );
    ErrorMessage( work1 );
    return;
  }

  GetString( 10 );	// Reading proc number
  if( sscanf( work,"%u",&procnum ) < 1 )
  {
    sprintf( work1,"Syntax error : '%s' is not a processor number",work );
    ErrorMessage( work1 );
    return;
  }

  if( ( pp = (PtProcessor)ProcessorList.Search( procnum ) ) == NULL )
  {
    sprintf( work1,"Invalid processor number : %u",procnum );
    ErrorMessage( work1 );
    return;
  };

  if( on == 1 )			// Type name if input node
  {
    GetString( 31 );
    if( *work == 0 )
    {
      sprintf( work1,"Syntax error : type name missing" );
      ErrorMessage( work1 );
      return;
    }
    strcpy( work1,work );
    GetString( 31 );
    if( sscanf( work,"%u",&us ) != 1 )
    {
      sprintf( work1,"Bad upsample factor : %s",work );
      ErrorMessage( work1 );
      return;
    }
    GetString( 31 );
    if( sscanf( work,"%u",&ds ) != 1 )
    {
      sprintf( work1,"Bad downsample factor : %s",work );
      ErrorMessage( work1 );
      return;
    }
  }
  else
	strcpy( work1,"TYPE1" );

  if( (ActBlock->NodeList).Search( node,OPTYPE_ALL ) != NULL )
  {
    sprintf( work1,"Node %u has already been defined",node );
    ErrorMessage( work1 );
    return;
  }
  if( node > MaxProbeNum )
	MaxProbeNum = node;
  p = new NodeInfo( node,OPTYPE_PROBE,in,on,ptype );
  (ActBlock->NodeList).Insert( p );
  p->SchedProc = (PtListObject)pp;
  in = ( in == 0 ? NOT_VALID : in );
  on = ( on == 0 ? NOT_VALID : on );
  np = new NodeObject( node,in,on );
  _fstrcpy( np->TypeName,work1 );
  if( in == NOT_VALID )		// Set only if input probe
  {
    np->UpScale = us;
    np->DownScale = ds;
  }
  p->GraphPtr = np;
  np->InfoNode = (PtListObject)p;

// If the node was output probe, an output NodeObject is also
// created. This hacking is done because an output probe
// has to have valid IPF else it won't get scheduled. IPFs
// are given only to output NodeObjects
  if( on == NOT_VALID )
  {
    np2 = new NodeObject( node, NOT_VALID, 1 );
    np2->InfoNode = (PtListObject)p;
    np->InsertOp( np2 );
  }
}

// Reads in a word. The word may be terminated with a space
// or with a closing parenthesis. Gets the character
// pointer and a pointer to the output buffer. Returns the
// new character pointer and the word in the output buffer
char *GetTypeWord( char *cp, char *buf )
{
  while( ( *cp != 0 ) && ( *cp != ' ' ) && ( *cp != ')' ) )
	*( buf++ ) = *( cp++ );
  *buf = 0;
  if( *cp == ' ' )
	++cp;
  return cp;
}

// Insert a node into the NodeInfo list. Evaluates
// the remaining part of the line and inserts the
// parameter list into the dynamic LISP database
void InsertNode()
{
  PtNodeInfo		p;
  PtNodeObject		np,ip;
  int			node,inps,outs;
  char			*cp;
  int			i,nesting,stat;
  char			type[32];
  PtBlock		bptr = NULL;
  PtPTrans		pptr;
  int			optype;

  GetString( 10 );	// Reading node number
  if( sscanf( work,"%u",&node ) < 1 )
  {
    sprintf( work1,"Syntax error : '%s' is not a node number",work );
    ErrorMessage( work1 );
    return;
  }

  GetString( 31 );		// Reading operator name
  if( *work == 0 )
  {
    sprintf( work1,"Syntax error : operator name missing" );
    ErrorMessage( work1 );
    return;
  }

  if( (ActBlock->NodeList).Search( node,OPTYPE_ALL ) != NULL )
  {
    sprintf( work1,"Node %u has already been defined",node );
    ErrorMessage( work1 );
    return;
  }

  bptr = ActBlock->SearchBlockName( work );
  if( bptr == NULL )	// No such block
  {
    sprintf( work1,"(get-io %s)\n",work );
    _fstrncpy( work2,DRCIExecuteCommand( CommWindow, work1 ),120);
    XDDeleteReply( CommWindow );
    if( sscanf( work2,"(%u %u)",&inps,&outs ) < 2 )
    {
      sprintf( work1,"Syntax error : %s is not a valid operator name",work );
      ErrorMessage( work1 );
      return;
    }
    SkipSpc();		// Ready to pass parameter list
    sprintf( work1,"(add-parlist %s '%s%u '%s\n",
		     work,ActBlock->BlockName,node,ActPtr );
    DRCIExecuteCommand( CommWindow, work1 );
    nesting = 0;
    stat = 0;
    cp = ActPtr;
    do
    {
      while( *cp )
      {
	if( *cp == '(' ) ++nesting;
	if( *cp == ')' ) --nesting;
	if( !nesting )
	{
	  stat = 1;
	  break;
	}
	++cp;
      }

      if( !stat )
      {
	if( ReadLine() )
	{
	  ErrorMessage( "Unexpected end of file" );
	  stat = 2;
	  FatalError();
	  return;
	}
	else
	{
	  sprintf( work1,"%s\n", ActPtr );
	  DRCIExecuteCommand( CommWindow, work1 );
	}
      }
    }
    while( !stat );
    DRCIExecuteCommand( CommWindow, ")\n" );
    XDDeleteReply( CommWindow );
  }
  else		// if( bptr == NULL )
  {
    inps = bptr->InputNum;
    outs = bptr->OutputNum;
  }

// Reading in operator type list
  if( bptr == NULL )
  {
    optype = OPTYPE_OP;
    sprintf( work1,"(get-optype %s)\n",work );
    _fstrncpy( work2,DRCIExecuteCommand( CommWindow, work1 ),120);
    XDDeleteReply( CommWindow );
    cp = work2+2;		// Skipping beginning parenthesises
  }
  else
  {
    IsDynamic = TRUE;	// Dynamic scheduling at root level
    optype = OPTYPE_BLOCK;
  }

  p = new NodeInfo( node,optype,inps,outs,work );
  (ActBlock->NodeList).Insert( p );

// Creating all the input nodes in advance so that
// IsReady() be able to detect unconnected input nodes
// The output nodes are also created because the
// operator procedure must always place its results
// somewhere. The variables created by the unconnected
// output nodes are cleared immediately.
  for( i = 1 ; i <= inps ; ++i )
  {
    if( bptr == NULL )
	cp = GetTypeWord( cp,type );
    else
    {
      pptr = (bptr->PTList).Search( i,PTRANS_INPUT );
      if( pptr == NULL )
      {
	sprintf( work1,"Input %u of block %s not found",
		 i,bptr->BlockName );
	ErrorMessage( work1 );
	return;
      }
      _fstrncpy( type,pptr->Type,31 );
    }
    np = new NodeObject( node,i,NOT_VALID );
// Propagating size data to the external node object
    if( bptr != NULL )
    {
      ip = BlockIntNode( i,PTRANS_INPUT,bptr );
      np->Size = ip->Size;
    }
    np->InfoNode = (PtListObject)p;
    _fstrcpy( np->TypeName,type );
    if( p->GraphPtr == NULL )
	p->GraphPtr = np;
    else
	(p->GraphPtr)->InsertOp( np );
  }
// Creating Control input NodeObject for a block
// if it has any
  if( ( bptr != NULL ) &&
      ( ( pptr = (bptr->PTList).Search( 0,PTRANS_INPUT ) ) != NULL ) )
  {
    np = new NodeObject( node,0,NOT_VALID );
    np->Size = 1L;
    _fstrncpy( type,pptr->Type,31 );
    np->InfoNode = (PtListObject)p;
    _fstrcpy( np->TypeName,type );
    if( p->GraphPtr == NULL )
	p->GraphPtr = np;
    else
	(p->GraphPtr)->InsertOp( np );
  };

// Looking for opening parenthesis of the output list
  if( bptr == NULL )
  {
    while( ( *cp != 0 ) && ( *cp != '(' ) )
	++cp;

    if( *cp == 0 )
    {
      sprintf( work,"Invalid type list was : %s",work2 );
      ErrorMessage( work );
      return;
    }
    else
	++cp;
  }

// Creating output node objects
  for( i = 1 ; i <= outs ; ++i )
  {
    if( bptr == NULL )
	cp = GetTypeWord( cp,type );
    else
    {
      pptr = (bptr->PTList).Search( i,PTRANS_OUTPUT );
      if( pptr == NULL )
      {
	sprintf( work1,"Output %u of block %s not found",
		 i,bptr->BlockName );
	ErrorMessage( work1 );
	return;
      }
      _fstrncpy( type,pptr->Type,31 );
    }
    np = new NodeObject( node,NOT_VALID,i );
// Propagating size data to the external node object
    if( bptr != NULL )
    {
      ip = BlockIntNode( i,PTRANS_OUTPUT,bptr );
      np->Size = ip->Size;
      np->Auto = ip->Auto;
    }
    np->InfoNode = (PtListObject)p;
    _fstrcpy( np->TypeName,type );
    if( p->GraphPtr == NULL )
	p->GraphPtr = np;
    else
	(p->GraphPtr)->InsertOp( np );
  }

  if( ( bptr == NULL ) && ( *cp != ')' ) )
  {
    sprintf( work,"Invalid type list was : %s",work2 );
    ErrorMessage( work );
    return;
  }
}

// Inserts a delay into the graph.
void InsertDelay()
{
  NODETYPE		node;
  PtNodeInfo		pi,po;
  PtDelayInputObject	npi;
  PtDelayOutputObject	npo;
  unsigned long		width;
  unsigned long		delay;

  GetString( 10 );	// Reading node number
  if( sscanf( work,"%u",&node ) < 1 )
  {
    sprintf( work1,"Syntax error : '%s' is not a node number",work );
    ErrorMessage( work1 );
    return;
  }

  GetString( 10 );	// Reading vector width
  if( sscanf( work,"%lu",&width ) < 1 )
  {
    sprintf( work1,"Syntax error : '%s' is not a delay width",work );
    ErrorMessage( work1 );
    return;
  }

  GetString( 10 );	// Reading delay length
  if( sscanf( work,"%lu",&delay ) < 1 )
  {
    sprintf( work1,"Syntax error : '%s' is not a delay length",work );
    ErrorMessage( work1 );
    return;
  }

  if( (ActBlock->NodeList).Search( node,OPTYPE_ALL ) != NULL )
  {
    sprintf( work1,"Node %u has already been defined",node );
    ErrorMessage( work1 );
    return;
  }

  pi = new NodeInfo( node,OPTYPE_IDELAY,1,1,"DELAY" );
  po = new NodeInfo( node,OPTYPE_ODELAY,1,1,"DELAY" );
  (ActBlock->NodeList).Insert( pi );
  (ActBlock->NodeList).Insert( po );
  npi = new DelayInputObject( node,1,NOT_VALID );
  npo = new DelayOutputObject( node,NOT_VALID,1 );
  pi->GraphPtr = (PtNodeObject)npi;
  po->GraphPtr = (PtNodeObject)npo;
  npi->InfoNode = (PtListObject)pi;
  npo->InfoNode = (PtListObject)po;
  npi->InsertOp( (PtNodeObject)npo );
  npo->Size = npi->Size = width;
  _fstrcpy( npi->TypeName,"TYPE1" );
  _fstrcpy( npo->TypeName,"TYPE1" );
  wsprintf( work1,"(add-parlist DELAY '%s%u '(%lu %lu))\n",
			(LPSTR)ActBlock->BlockName,
			node,width,delay );
  DRCIExecuteCommand( CommWindow, work1 );
  XDDeleteReply( CommWindow );
}

// Inserts a connection into the graph
// Evaluates the remaining parts of the line
void InsertConnection()
{
  INPUTTYPE	InputNode,InputNum;
  INPUTTYPE	OutputNode,OutputNum;
  BOOL		Create = FALSE;
  char		ac;
  PtNodeInfo	OutInfoPtr;
  PtNodeInfo	InInfoPtr;
  PtNodeObject	ONP;
  PtNodeObject	INP;
  unsigned long	isize,osize;
  BOOL		iset,oset;

  GetString( 10 );
  if( sscanf( work,"%u_%u",&OutputNode,&OutputNum ) < 2 )
  {
    sprintf( work1,"OUTCONN field invalid %s",work );
    ErrorMessage( work1 );
    return;
  }

  GetString( 10 );
  if( sscanf( work,"%u_%u",&InputNode,&InputNum ) < 2 )
    if( sscanf( work,"%u_C",&InputNode ) == 1 )
    {
      InputNum = 0;  // Condition entry
      isize = 1L;    // Default, can be overwritten !
    }
    else
    {
      sprintf( work1,"INCONN field invalid %s",work );
      ErrorMessage( work1 );
      return;
    }

  if( ( OutInfoPtr = (PtNodeInfo)(ActBlock->NodeList).Search( OutputNode,OPTYPE_OUT ) ) == NULL )
  {
    sprintf( work1,"Reference to not defined node : %u",OutputNode );
    ErrorMessage( work1 );
    return;
  }

  if( OutputNum > OutInfoPtr->Outputs )
  {
    sprintf( work1,"Node %u has no output %u",OutputNode,OutputNum );
    ErrorMessage( work1 );
    return;
  }

// Examines if the output node must be created
// Anything exists from the node chain ?
  if( OutInfoPtr->GraphPtr == NULL )
	Create = TRUE;
  else
// This particuliar output exists ?
  if( ( ONP = (OutInfoPtr->GraphPtr)->Search( NOT_VALID,OutputNum ) ) == NULL )
	Create = TRUE;

// Node object must be created for this output
  if( Create )
  {
    ONP = new NodeObject( OutputNode,NOT_VALID,OutputNum );
    ONP->InfoNode = (PtListObject)OutInfoPtr;
    if( OutInfoPtr->GraphPtr == NULL )
	OutInfoPtr->GraphPtr = ONP;
    else
	(OutInfoPtr->GraphPtr)->InsertOp( ONP );
  }

// Now comes the input
  if( ( InInfoPtr = (PtNodeInfo)(ActBlock->NodeList).Search( InputNode,OPTYPE_IN ) )
		     == NULL )
  {
    sprintf( work1,"Reference to not defined node : %u",InputNode );
    ErrorMessage( work1 );
    return;
  }

// Examines if the input node must be created
// Anything exists from the node chain ?
  if( InInfoPtr->GraphPtr == NULL )
  {
    sprintf( work1,"Node %u has no input",InputNode );
    ErrorMessage( work1 );
    return;
  }

  INP = (InInfoPtr->GraphPtr)->Search( InputNum,NOT_VALID );
  if( INP == NULL )
  {
    sprintf( work1,"Node %u has no input %u",InputNode,InputNum );
    ErrorMessage( work1 );
    return;
  }

  if( INP->UpNode != NULL )
  {
    sprintf( work1,"An output already connected to node %u input %u",InputNode,InputNum );
    ErrorMessage( work1 );
    return;
  }

  if( ( OutInfoPtr->Type == OPTYPE_PROBE ) &&
      ( InInfoPtr->Type == OPTYPE_PROBE ) )
  {
    sprintf( work1,"Node %u and %u are both probes",InputNode,OutputNode );
    ErrorMessage( work1 );
    return;
  }

// Filling the Size field of the NodeObjects and checking
// if they are equal
  iset = oset = FALSE;
  if( InInfoPtr->Type &
	( OPTYPE_OP | OPTYPE_IDELAY | OPTYPE_ODELAY ) )
  {
    wsprintf( work,"(get-isize %s %u '%s%u)\n",
		(LPSTR)InInfoPtr->Operator,
		InputNum,
		(LPSTR)ActBlock->BlockName,
		InputNode );
    _fstrcpy( work,DRCIExecuteCommand( CommWindow, work ) );
    XDDeleteReply( CommWindow );
    if( sscanf( work,"%lu",&isize ) < 1 )
    {
      ErrorMessage( work );
      return;
    }
    iset = TRUE;
  }
  else
  if( InInfoPtr->Type & OPTYPE_BLOCK )
  {
    isize = INP->Size;
    iset = TRUE;
  }

  if( OutInfoPtr->Type &
		( OPTYPE_OP | OPTYPE_IDELAY | OPTYPE_ODELAY ) )
  {
    if( ONP->Size != 0L )
	osize = ONP->Size;
    else
    {
      wsprintf( work,"(get-osize %s %u '%s%u)\n",
		(LPSTR)OutInfoPtr->Operator ,
		OutputNum,
		(LPSTR)ActBlock->BlockName,
		OutputNode );
      _fstrcpy( work,DRCIExecuteCommand( CommWindow, work ) );
      XDDeleteReply( CommWindow );
      if( sscanf( work,"(%lu %c)",&osize,&ac ) < 2 )
      {
	ErrorMessage( work );
	return;
      }
      if( ac != 'A' )
		ONP->Auto = FALSE;
    }
    oset = TRUE;
  }
  else
  if( OutInfoPtr->Type & OPTYPE_BLOCK )
  {
    osize = ONP->Size;
    oset = TRUE;
  }

// Set sizes
  if( iset & !oset )
	osize = isize;
  else
  if( !iset & oset )
	isize = osize;
  else
  if( !iset & !oset )
  {
    sprintf( work1,"The sizes of %u_%u and %u_%u are both undefined",
		   InputNode,InputNum,OutputNode,OutputNum );
    ErrorMessage( work1 );
    return;
  }

  if( iset & oset & ( isize != osize ) )
  {
    sprintf( work1,"%u_%u is not the same size as %u_%u (I:%u,O:%u)",
	     InputNode,InputNum,OutputNode,OutputNum,isize,osize );
    ErrorMessage( work1 );
    return;
  }
  INP->Size = isize;
  ONP->Size = osize;

// Inserting the connection into the graph
  ONP->Insert( INP );
  if( InInfoPtr->Type == OPTYPE_PROBE )
  {
    wsprintf( work,"(add-parlist INPPROBE '%s%u '(%lu))\n",
			(LPSTR)ActBlock->BlockName,
			InInfoPtr->NodeNum,isize );
    DRCIExecuteCommand( CommWindow, work );
    XDDeleteReply( CommWindow );
  }

  if( OutInfoPtr->Type == OPTYPE_PROBE )
  {
    wsprintf( work,"(add-parlist OUTPROBE '%s%u '(%lu))\n",
			(LPSTR)ActBlock->BlockName,
			OutInfoPtr->NodeNum,isize );
    DRCIExecuteCommand( CommWindow, work );
    XDDeleteReply( CommWindow );
  }
}

// Creates the channel and activity sublist in a block
// descriptor.
void BuildChActList()
{
  int	i;
  PtChannel CP;
  PtProcessor	p;
  PtActivity	ActivityItem;
  PtVariable	VariableItem;
  PtChunk	ChunkItem;

  p = (PtProcessor)ProcessorList.SubList;
  while( p != NULL )
  {
    ActivityItem = new Activity( p->NodeNum,0,0L,0 );
    (ActBlock->ActivityList).Insert( ActivityItem );
    ChunkItem = new Chunk( p->NodeNum,0L,0L,p->Memory );
    (ActBlock->ChunkList).Insert( ChunkItem );
    VariableItem = new Variable( p->NodeNum,0L );
    (ActBlock->VariableList).Insert( VariableItem );
    p = (PtProcessor)p->BackLink;
  }

  for( i = 1 ; i <= NumberOfChannels ; ++i )
  {
    CP = new Channel( i );
    (ActBlock->Channels).Insert( CP );
  }
}

// Inserts a block descriptor into the graph
void InsertBlock()
{
  int	Inum = 0, Onum = 0;
  int	s,se,n;
  PtBlock	BPtr;
  PtPTrans	TPtr;
  BOOL		f = TRUE;
  char		tname[32];

  if( !ActBlock->IsRoot )
  {
    sprintf( work1,"Block nesting is forbidden" );
    ErrorMessage( work1 );
    FatalError();
  }

  GetString( 32 );
  if( !( *work ) )
  {
    sprintf( work1,"Block name missing" );
    ErrorMessage( work1 );
    return;
  }
  BPtr = new Block( work,FALSE );
  do
  {
    GetString( 48 );
    if( *work == 0 )
	f =  FALSE;
    else
    if( sscanf( work,"C:%s",tname ) == 1 )
    {
      TPtr = new PTrans( 0,0,0,PTRANS_INPUT,tname );
      (BPtr->PTList).Insert( TPtr );
    }
    else
    if( sscanf( work,"I%u->%u_%u:%s",&s,&n,&se,tname ) == 4 )
    {
      TPtr = new PTrans( s,n,se,PTRANS_INPUT,tname );
      (BPtr->PTList).Insert( TPtr );
      if( s > Inum )
	Inum = s;
    }
    else
    if( sscanf( work,"O%u->%u_%u:%s",&s,&n,&se,tname ) == 4 )
    {
      TPtr = new PTrans( s,n,se,PTRANS_OUTPUT,tname );
      (BPtr->PTList).Insert( TPtr );
      if( s > Onum )
	Onum = s;
    }
    else
    {
      sprintf( work1,"Invalid I/O qualifier : %s",work );
      ErrorMessage( work1 );
      return;
    }
  }
  while( f );
  BPtr->InputNum = Inum;
  BPtr->OutputNum = Onum;
  ActBlock->Insert( BPtr );
  ActBlock = BPtr;
  BuildChActList();
}

// Closes the definition of a block
void CloseBlock()
{
  PtPTrans	pt;
  PtNodeInfo	InfoPtr;
  PtNodeObject	ip;
  unsigned long	isize;
  char		aflag;

  GetString( 32 );
  if( !( *work ) )
  {
    sprintf( work,"Block name missing" );
    ErrorMessage( work );
    return;
  }
  if( _fstrncmp( work,ActBlock->BlockName,31 ) != 0 )
  {
    sprintf( work1,"%s does not match %s",
	     work,ActBlock->BlockName );
    ErrorMessage( work1 );
    return;
  }
  if( ActBlock->IsRoot )
  {
    sprintf( work,"Cannot close ROOT" );
    ErrorMessage( work );
    return;
  }

// Fetching internal node input and output sizes and
// propagating them to outside
  pt = (PtPTrans)(ActBlock->PTList).SubList;
  while( pt != NULL )
  {
    if( pt->NodeNum != 0 ) // If not control input
    {
      if( pt->Dir == PTRANS_INPUT )
      {
	ip = BlockIntNode( pt->NodeNum,PTRANS_INPUT,ActBlock );
	InfoPtr = (ActBlock->NodeList).Search( pt->PNodeNum,OPTYPE_ALL );
	wsprintf( work,"(get-isize %s %u '%s%u)\n",
		(LPSTR)InfoPtr->Operator,
		pt->PIONum,
		(LPSTR)ActBlock->BlockName,
		pt->PNodeNum );
	_fstrcpy( work,DRCIExecuteCommand( CommWindow, work ) );
	XDDeleteReply( CommWindow );
	if( sscanf( work,"%lu",&isize ) < 1 )
	{
	  wsprintf( work1,"CloseBlock.get-isize : %s",
		  (LPSTR)work );
	  FatalError( work1 );
	  return;
	}
	ip->Size = isize;
      }
      else
      {
	ip = BlockIntNode( pt->NodeNum,PTRANS_OUTPUT,ActBlock );
	InfoPtr = (ActBlock->NodeList).Search( pt->PNodeNum,OPTYPE_ALL );
	wsprintf( work,"(get-osize %s %u '%s%u)\n",
		(LPSTR)InfoPtr->Operator,
		pt->PIONum,
		(LPSTR)ActBlock->BlockName,
		pt->PNodeNum );
	_fstrcpy( work,DRCIExecuteCommand( CommWindow, work ) );
	XDDeleteReply( CommWindow );
	if( sscanf( work,"(%lu %c)",&isize,&aflag ) < 2 )
	{
	  wsprintf( work1,"CloseBlock.get-osize : %s",
		  (LPSTR)work );
	  FatalError( work1 );
	  return;
	}
	ip->Size = isize;
	if( aflag == 'A' )
	  ip->Auto = TRUE;
      }
    }
    pt = (PtPTrans)pt->BackLink;
  }
  ActBlock = &RootBlock;
}

// Reads in the node descriptor part of the input file
void ReadNodes()
{
  BOOL	f = TRUE,EmptyLine;

  while( f )
  {
    f = !ReadLine();
    SkipSpc();
    EmptyLine = *ActPtr == 0;
    if( f && !EmptyLine ) // If line was read
    {
      if( fit( ";" ) )
		continue;
      else
      if( fit( "CONNECTIONS" ) )
			continue; // to remain compatible
      else
      if( fit( "PROBE" ) )
		InsertProbe();
      else
      if( fit( "NODE" ) )
		InsertNode();
      else
      if( fit( "DELAY" ) )
		InsertDelay();
      else
      if( fit( "CONNECTION" ) )
		InsertConnection();
      else
      if( fit( "BLOCK" ) )
		InsertBlock();
      else
      if( fit( "ENDBLOCK" ) )
		CloseBlock();
      else
      {
	sprintf( work1,"Invalid keyword : %s",GetString(20));
	ErrorMessage( work1 );
      }
    }
  }
}

void DescErrorMessage( char msg[] )
{
  char	work[140];

  IncError();
  sprintf( work,"Line %lu in SYSTEM.DSC : %s",LineCtr,msg );
  SendMessageLine( work );
}

// Evaluates a processor definition
void InsertProcessor()
{
  unsigned	Procnum;
  char		Dev[32];
  unsigned long	Mem;
  PtProcessor	NP;

  SkipSpc();
  GetString( 4 );
  if( sscanf( work,"%u",&Procnum ) < 1 )
  {
    sprintf( work1,"Cannot evaluate proc num : '%s'",work );
    DescErrorMessage( work1 );
    return;
  }
  SkipSpc();
  GetString( 31 );
  if( *work == 0 )
  {
    DescErrorMessage( "Device name not found" );
    return;
  }
  strcpy( Dev,work );
  SkipSpc();
  GetString( 10 );
  if( sscanf( work,"%lu",&Mem ) < 1 )
  {
    sprintf( work1,"Cannot evaluate memory size : '%s'",work );
    DescErrorMessage( work1 );
    return;
  }
  if( ProcessorList.Search( Procnum ) != NULL )
  {
    sprintf( work1,"Processor %u already declared",Procnum );
    DescErrorMessage( work1 );
    return;
  }
}

// Reads the system descriptor information from the database
void ReadSysDesc()
{
  BOOL		f = TRUE;
  unsigned long Mem;
  char		Dev[32];
  unsigned	ProcNum;
  PtProcessor	NP;

  do
  {
    wsprintf( work,"(get-nextproc)\n" );
    _fstrcpy( work,DRCIExecuteCommand( CommWindow, work ) );
    XDDeleteReply( CommWindow );
    if( sscanf( work,"(%u %s %lu)",&ProcNum,Dev,&Mem ) < 3 )
	f = FALSE;
    else
    {
      NP = new Processor( ProcNum,Mem,Dev );
      ProcessorList.Insert( NP );
    }
  }
  while( f );
}

// Asks the database about the number of communication
// channels in the system
void AskChannelNum()
{
  wsprintf( work,"(get-channelnum)\n" );
  _fstrcpy( work,DRCIExecuteCommand( CommWindow, work ) );
  XDDeleteReply( CommWindow );
  if( sscanf( work,"%u",&NumberOfChannels ) < 1 )
  {
    sprintf( work1,"get-channelnum returned : %s",work );
    FatalError( work1 );
  }
}

// Returns TRUE if the type name is a dynamic one.
BOOL IsDynamicTypeName( LPSTR name )
{
  int	n;
  char	Type[32];

  _fstrcpy( Type,name );
  return( _fstrncmp( Type,"TYPE",4 ) == 0 ? TRUE : FALSE );
}

// Checks the type names and substitutes if necessary
void CheckTypeNames( PtNodeObject np, LPSTR ParentName )
{
  char	Name[32];

  if( !IsDynamicTypeName( ParentName ) )
  {
    if( IsDynamicTypeName( np->TypeName ) )
    {
      _fstrcpy( Name,np->TypeName );
      np->SubstituteTypeName( Name,ParentName );
    }
    else
    if( _fstrcmp( ParentName,np->TypeName ) != 0 )
    {
      if( np->InputNum != NOT_VALID )
	wsprintf( work1,"Type conflict at node %u input %u between %s and %s",
			np->NodeNum,np->InputNum,
			ParentName,(LPSTR)np->TypeName );
      else
	wsprintf( work1,"Type conflict at node %u output %u between %s and %s",
			np->NodeNum,np->OutputNum,
			ParentName,(LPSTR)np->TypeName );
      Error2Message( work1 );
    }
  }
}

// Changes upscale value by factor. Gets pointers to the
// upscale and downscale values and the factor to
// change with
void ChangeUpScale( unsigned *usp, unsigned *dsp, unsigned factor )
{
   if( ( *dsp % factor ) == 0 )
	*dsp /= factor;
   else
	*usp *= factor;
}

// Changes downscale value by factor. Gets pointers to the
// upscale and downscale values and the factor to
// change with
void ChangeDownScale( unsigned *usp, unsigned *dsp, unsigned factor )
{
   if( ( *usp % factor ) == 0 )
	*usp /= factor;
   else
	*dsp *= factor;
}

// Propagates the IPF into the embedded node of a block
// Gets pointer to the external node of the block, its
// info pointer and the block pointer
void PropagateIPF( PtNodeInfo ip, PtNodeObject np,
		   PtBlock bptr )
{
  PtNodeObject	nodeptr;

  if( np->InputNum != 0 )
  {
    nodeptr = BlockIntNode( np->InputNum, PTRANS_INPUT,bptr );
    nodeptr->UpScale = np->UpScale;
    nodeptr->DownScale = np->DownScale;
    nodeptr->Ready = TRUE;
  }
}

void CheckNodeIPF( PtNodeObject, unsigned, unsigned);

// Checks the block's IPF factors. Gets the Infoptr of the
// block and calculates the IPF for the embedded operators
// Returns FALSE if the block cannot be evaluated because
// not all the inputs has been visited yet
BOOL CheckBlockIPF( PtNodeInfo InfoPtr,unsigned *ui,
		    unsigned *di )
{
  PtBlock	bptr,sbptr;
  PtNodeObject	np,np2;
  PtPTrans	tp;

  bptr = ActBlock->SearchBlockName( InfoPtr->Operator );
// First we check whether all the input nodes has been
// visited and return FALSE if not
  np = InfoPtr->GraphPtr;
  while( np != NULL )
  {
    if( (np->InputNum != NOT_VALID) && !np->Ready )
	return FALSE;
    if( np->InputNum != NOT_VALID )
	PropagateIPF( InfoPtr,np,bptr );
    np = np->ForwNode;
  }

  np = InfoPtr->GraphPtr;
  while( np != NULL )
  {
    if( (np->InputNum != NOT_VALID ) && !np->Ready )
	return FALSE;
    if( np->InputNum != NOT_VALID )
	PropagateIPF( InfoPtr,np,bptr );
    np = np->BackNode;
  }
  sbptr = ActBlock;
  ActBlock = bptr;
// Checks node IPF for every node input
  tp = (PtPTrans)(bptr->PTList).SubList;
  while( tp != NULL )
  {
    if( ( tp->Dir == PTRANS_INPUT ) &&
	( tp->NodeNum != 0 ) )
    {
      np = BlockIntNode( tp->PIONum,PTRANS_INPUT,bptr);
      CheckNodeIPF( np, np->UpScale, np->DownScale );
    }
    tp = (PtPTrans)tp->BackLink;
  }
// propagates embedded node IPF to outside
  tp = (PtPTrans)(bptr->PTList).SubList;
  *ui = 0;
  *di = 0;
  while( tp != NULL )
  {
    if( tp->Dir == PTRANS_OUTPUT )
    {
      np = BlockIntNode( tp->PIONum,PTRANS_OUTPUT,bptr );
      np2 = (InfoPtr->GraphPtr)->Search( NOT_VALID,tp->PIONum );
      np2->UpScale = np->UpScale;
      np2->Ready = TRUE;
      if( *ui == 0 )
	*ui = np2->UpScale;
      else
      if( *ui != np2->UpScale )
      {
	wsprintf( work1,"outputs have different IPF : output %u",
		  np2->NodeNum );
	Error2Message( work1 );
      }
      np2->DownScale = np->DownScale;
      if( *di == 0 )
	*di = np2->DownScale;
      else
      if( *di != np2->DownScale )
      {
	wsprintf( work1,"outputs have different IPF : output %u",
		  np2->NodeNum );
	Error2Message( work1 );
      }
    }
    tp = (PtPTrans)tp->BackLink;
  }
  ActBlock = sbptr;
  return TRUE;
}

// Sets Visited variable to TRUE of all the InfoNodes
// belonging to the nodeobject group. Generally only
// one InfoNode belongs to the same nodeobject group
// but in the case of delays there's two.
void VisitNode( PtNodeObject np )
{
  PtNodeObject	np2;
  PtNodeInfo    ip;

  np2 = np;
  while( np2 != NULL )
  {
    ip = (PtNodeInfo)(np2->InfoNode);
    ip->Visited = TRUE;
    np2 = np2->ForwNode;
  }
  np2 = np;
  while( np2 != NULL )
  {
    ip = (PtNodeInfo)(np2->InfoNode);
    ip->Visited = TRUE;
    np2 = np2->BackNode;
  }
}

// Asks the IPFs if necessary and sets it on the given output
// node. Return FALSE if the node IPF could not be  evaluated.
// It can happen if the node is a block whose all inputs
// has not been visited.
BOOL AskIPF( PtNodeObject np2 )
{
  PtNodeInfo	ip;
  PtNodeObject	np;
  unsigned	wus,wds;
  BOOL		nfound = FALSE;
  unsigned	wuipf,wdipf;

  if( np2 == NULL )
	return FALSE;
  ip = (PtNodeInfo)np2->InfoNode;
  if( np2->UIPF == 0 )
  {
    if( ip->Type & OPTYPE_BLOCK )
    {
      if( !(((PtNodeInfo)( np2->InfoNode ))->Visited) )
      {
	if( !CheckBlockIPF( ip,&wus,&wds ) )
		return FALSE;
      }
      else
      {
	wus = np2->UpScale;
	wds = np2->DownScale;
      }
      wuipf = wdipf = 1;
      np = np2;
      while( ( np != NULL ) && !nfound )
      {
	if( np->InputNum != NOT_VALID )
	{
	  wuipf = np->UpScale;
	  wdipf = np->DownScale;
	  nfound = TRUE;
	}
	np = np->ForwNode;
      }
      if( !nfound )
		np = np2;
      while( ( np != NULL ) && !nfound )
      {
	if( np->InputNum != NOT_VALID )
	{
	  wuipf = np->UpScale;
	  wdipf = np->DownScale;
	  nfound = TRUE;
	}
	np = np->BackNode;
      }
      np2->UIPF = wus / wuipf;
      np2->DIPF = wds / wdipf;
      VisitNode( np2 );
    }
    else
    {
      wsprintf( work,"(get-IPF %s %u '%s%u)\n",
	     (LPSTR)ip->Operator, np2->OutputNum,
	     (LPSTR)ActBlock->BlockName,
	     np2->NodeNum );
      _fstrncpy( work,DRCIExecuteCommand( CommWindow, work ),120);
      XDDeleteReply( CommWindow );
      if( sscanf( work,"(%u %u)",&wus,&wds ) < 2 )
      {
	sprintf( work2,"XLisp error : %s",work );
	ErrorMessage( work2 );
	FatalError();
      }
      np2->UIPF = wus;
      np2->DIPF = wds;
      VisitNode( np2 );
    }
  }
  return TRUE;
}

// Sets the IPF fields of an output node provided that
// the IPF is propagated through that node
void SetForwardIPF( PtNodeObject node, unsigned us,unsigned ds )
{
  ChangeUpScale( &us,&ds,node->UIPF );
  ChangeDownScale( &us,&ds,node->DIPF );
  node->UpScale = us;
  node->DownScale = ds;
}

// Verifies the IPFs on a node and sends error messages
// if discrepancies were found
void VerifyIPF( PtNodeObject np2, unsigned us, unsigned ds )
{
  float fac;
  char	msg1[20];
  unsigned	inum;

  if( np2->UpScale == 0 )
	np2->UpScale = us;
  else
  if( us != np2->UpScale )
  {
    if( np2->InputNum == NOT_VALID )
    {
      strcpy( msg1,"output" );
      inum = np2->OutputNum;
    }
    else
    {
      strcpy( msg1,"input" );
      inum = np2->InputNum;
    }
    sprintf( work,"IPF violation at node %u %s %u",
	     np2->NodeNum, msg1 , inum );
    ErrorMessage( work );
    return;
  }

  if( np2->DownScale == 0 )
	np2->DownScale = ds;
  else
  if( ds != np2->DownScale )
  {
    if( np2->InputNum == NOT_VALID )
    {
      strcpy( msg1,"output" );
      inum = np2->OutputNum;
    }
    else
    {
      strcpy( msg1,"input" );
      inum = np2->InputNum;
    }
    sprintf( work,"IPF violation at node %u %s %u",
	     np2->NodeNum, msg1 , inum );
    ErrorMessage( work );
    return;
  }
  fac = ((float)np2->DownScale)/((float)np2->UpScale);
  if( fac > LFactor )
  {
    LFactor = fac;
    LUpScale = np2->UpScale;
    LDownScale = np2->DownScale;
  }
}

// Check one node IPF and propagates it through its
// subgraph. Signals error message if IPF violations
// were found.
void CheckNodeIPF( PtNodeObject node, unsigned us, unsigned ds )
{
  PtNodeObject	np2,np3,np4;
  unsigned	wus,wds;

  fprintf( DFile,"Block %s,Node %u,I:%u,O:%u->%u,%u\n",
		 ActBlock->BlockName,node->NodeNum,
		 node->InputNum,node->OutputNum,
		 us,ds );

// For safety
  if( node == NULL )
	return;
  np3 = (PtNodeObject)node;
  while( np3 != NULL )
  {
    if( np3->OutputNum != NOT_VALID )	// if output
    {
      if( AskIPF( np3 ) )
      {
	SetForwardIPF( np3,us,ds );
	np3->Ready = TRUE;
	np2 = (PtNodeObject)np3->SubList;
	while( np2 != NULL )
	{
	  VerifyIPF( np2,np3->UpScale,np3->DownScale );
	  if( !(((PtNodeInfo)(np2->InfoNode))->Visited) )
	  {
	    np2->Ready = TRUE;
	    CheckNodeIPF( np2,np3->UpScale,np3->DownScale );
	  }
	  np2 = (PtNodeObject)np2->BackLink;
	}
      }
    }
    else		// if input
    {
      VerifyIPF( np3,us,ds );
      np4 = np3->UpNode;
      np3->Ready = TRUE;
      if( np4 != NULL )
      {
	VerifyIPF( np4,us,ds );
	if( !np4->Ready )
	{
	  if( AskIPF( np4 ) )
	  {
	    np4->Ready = TRUE;
	    wus = us;
	    wds = ds;
	    ChangeUpScale( &wus,&wds,np4->DIPF );
	    ChangeDownScale( &wus,&wds,np4->UIPF );
	    CheckNodeIPF( np4,wus,wds );
	  }
	}
      }
    }
    np3 = (PtNodeObject)np3->ForwNode;
  }
  np3 = (PtNodeObject)node->BackNode;
  while( np3 != NULL )
  {
    if( np3->OutputNum != NOT_VALID )	// if output
    {
      if( AskIPF( np3 ) )
      {
	SetForwardIPF( np3,us,ds );
	np3->Ready = TRUE;
	np2 = (PtNodeObject)np3->SubList;
	while( np2 != NULL )
	{
	  VerifyIPF( np2,np3->UpScale,np3->DownScale );
	  if( !(((PtNodeInfo)(np2->InfoNode))->Visited) )
	  {
	    np2->Ready = TRUE;
	    CheckNodeIPF( np2,np3->UpScale,np3->DownScale );
	  }
	  np2 = (PtNodeObject)np2->BackLink;
	}
      }
    }
    else		// if input
    {
      VerifyIPF( np3,us,ds );
      np4 = np3->UpNode;
      np3->Ready = TRUE;
      if( np4 != NULL )
      {
	VerifyIPF( np4,us,ds );
	if( !np4->Ready )
	{
	  if( AskIPF( np4 ) )
	  {
	    np4->Ready = TRUE;
	    wus = us;
	    wds = ds;
	    ChangeUpScale( &wus,&wds,np4->DIPF );
	    ChangeDownScale( &wus,&wds,np4->UIPF );
	    CheckNodeIPF( np4,wus,wds );
	  }
	}
      }
    }
    np3 = (PtNodeObject)np3->BackNode;
  }
}

// Checks interpolation factors by propagating the IPFs
// from the input probes. Fills in the node IPF fields
// and signals IPF errors
void CheckIPF()
{
  PtNodeInfo	p;
  PtNodeObject	np,np2;

  ResetAllNodeFlags();
  LFactor = -1;
  p = (PtNodeInfo)(ActBlock->NodeList).SubList;
  DFile = fopen( "rcomp.ipd","w" );
  while( p != NULL )
  {
// If input probe
    if( ( p->Type == OPTYPE_PROBE ) && ( p->GraphPtr != NULL )
	&& ( p->Inputs == 0 ) )
    {
      np = p->GraphPtr;
      CheckNodeIPF( np,np->UpScale,np->DownScale );
    }
    p = (PtNodeInfo)p->BackLink;
  }
  fclose( DFile );
}

// Checks the node type against the type of its parent
//		***** WARNING : RECURSIVE *****
// Gets the node to be checked and the type name of its parent
void CheckNodeType( PtNodeObject np, LPSTR ParentName )
{
  PtNodeObject  np2;
  PtNodeObject	np3;
  char		Name[32];

  CheckTypeNames( np,ParentName );
  np2 = np->BackNode;
  while( np2 != NULL )
  {
    if( !IsDynamicTypeName( np2->TypeName ) )
      if( np2->InputNum != NOT_VALID )	// If input node,check its parent
	CheckTypeNames( np2->UpNode,np2->TypeName );
      else
      {
	np3 = (PtNodeObject)np2->SubList;
	_fstrcpy( Name,np2->TypeName );
	while( np3 != NULL )
	{
	  CheckNodeType( np3,Name );
	  np3 = (PtNodeObject)np3->BackLink;
	}
      }
    np2 = np2->BackNode;
  }

  np2 = np->ForwNode;
  while( np2 != NULL )
  {
    if( !IsDynamicTypeName( np2->TypeName ) )
      if( np2->InputNum != NOT_VALID )	// If input node,check its parent
	CheckTypeNames( np2->UpNode,np2->TypeName );
      else
      {
	np3 = (PtNodeObject)np2->SubList;
	_fstrcpy( Name,np2->TypeName );
	while( np3 != NULL )
	{
	  CheckNodeType( np3,Name );
	  np3 = (PtNodeObject)np3->BackLink;
	}
      }
    np2 = np2->ForwNode;
  }
}

// Type checking routine. Spreads the statical type names
// accross the graph and finds out if there are type
// mismatches
void CheckTypes()
{
  PtNodeInfo	p;
  PtNodeInfo	p1;
  PtNodeObject	np;
  NODETYPE	nodenum;
  char		probename[32];

  p = (PtNodeInfo)(ActBlock->NodeList).SubList;
  while( p != NULL )
  {
// If probe
    if( ( p->Type == OPTYPE_PROBE ) && ( p->GraphPtr != NULL ) )
    {
      _fstrcpy( probename,(p->GraphPtr)->TypeName );
      np = (PtNodeObject)(p->GraphPtr)->SubList;
      while( np != NULL )
      {
	CheckNodeType( np,probename );
	np = (PtNodeObject)np->BackLink;
      }
    }
    p = (PtNodeInfo)p->BackLink;
  }
}

// Compiling routine. Supposes SrcFile,MsgFile to
// be open.
void rcomp()
{

// Initializing section
  ActBlock = &RootBlock;
  IsDynamic = FALSE;
  LineCtr = 0L;
  ErrCtr = 0;
  WarnCtr = 0;
  MaxProbeNum = 0;
  VarID = 0L;
  ErrorWithLineNumbers = TRUE;
  DRCIExecuteCommand( CommWindow, "(init-comp)\n" );
  XDDeleteReply( CommWindow );
  UpdateDlgText( MD_STATUSTEXT,"Reading target description ..." );

// Reads in the hardware description info
   ReadSysDesc();

// Creates processor descriptors for Activity,Chunk and
// Variable and Channel lists
  AskChannelNum();
  BuildChActList();

// Translating section
  LineCtr = 0;
  if( ErrCtr == 0 )
  {
    UpdateDlgText( MD_STATUSTEXT,"Reading SFG nodes ..." );
    ReadNodes();
  }

  ErrorWithLineNumbers = FALSE;
  if( ErrCtr == 0 )
  {
    UpdateDlgText( MD_STATUSTEXT,"Type checking ..." );
    CheckTypes();
  }

  if( ErrCtr == 0 )
  {
    UpdateDlgText( MD_STATUSTEXT,"IPF checking ..." );
    CheckIPF();
  }

  ResetAllNodeFlags();

  if( ErrCtr == 0 )
  {
    UpdateDlgText( MD_STATUSTEXT,"Scheduling ..." );
    ScheduleBlockTree();
  }

  if( ErrCtr == 0 )
  {
    UpdateDlgText( MD_STATUSTEXT,"Generating output file ..." );
    Codegen();
  }

// Cleanup section
  DRCIExecuteCommand( CommWindow, "(terminate-comp)\n" );
  XDDeleteReply( CommWindow );
}
