/******** rc2.cpp
		Rafael compiler,part 2
		Written by Gabor Paller
				******/

#include	<windows.h>
#include	<stdio.h>
#include	<ctype.h>
#include	<string.h>
#include	"robject.h"
#include	"rcomp.h"
#include	"xserver.h"
#include	"rci.h"

// Structure used for processor pair-channel number cache
#define CHANNELCACHE struct channelcache
#define	CCACHESIZE	16
CHANNELCACHE
{
  int	FromProc;
  int	ToProc;
  int	ChannelNum;
};

int		CCachePtr;	// Points to next free element in the cache
CHANNELCACHE	ccache[CCACHESIZE];	// Channel cache

// Dictionary of nodes
NodeInfo	ReadyList( 0,0,0,0,"" );

extern PtBlock		ActBlock;
extern Block		RootBlock;
PtNodeInfo		ActBlockNodeInfo;
extern Processor	ProcessorList;
extern Variable		VariableList;
ExecTimeList		TimeList( "",0 );

extern INPUTTYPE	MaxProbeNum;

// TRUE if dynamic scheduling at root level
extern BOOL	IsDynamic;

// Message file
extern FILE		*MsgFile;
// Line counter
extern unsigned long	LineCtr;
// Error counter
extern unsigned ErrCtr;
// Warning counter
extern unsigned WarnCtr;
// Handle of the main window
extern HWND MainWindow;
// Handle of the comm window
extern HWND CommWindow;

// Variable ID counter
extern unsigned long VarID;

// UpScale and DownScale limit values from RC1
extern unsigned LUpScale,LDownScale;

// work string variable
extern char work[1024];
// work string variable 2
extern char work1[140];
// work string variable 3
extern char work2[1024];
// device list variable
char	DeviceList[1024];
// Virtual Time
unsigned long	VirtualTime;
// Actual UpScale and DownScale factors
unsigned 	UpScale,DownScale;

// Weighting constants
float	MemReqWeight;		// Memory requirement weight
float	FreedomWeight;		// Freedom weight

// Imported functions from RCOMP.CPP
void SendMessageLine( char * );
void UpdateDlgText( int,char * );
void FatalError();
void FatalError( char[] );
LPSTR DRCIExecuteCommand( HWND,LPSTR );

// Imported functions from RC1.CPP
void WarningMessage( char[] );
void ErrorMessage( char[] );
void Error2Message( char[] );

unsigned long maxl( unsigned long a, unsigned long b )
{
  return( a < b ? b : a );
}

// Creates the list of all the devices into the DeviceList
// variable. Each device is shown only once
void CreateDeviceList()
{
  PtProcessor	pp;
  BOOL		Found = FALSE;
  char		*cp,*cp1;
  char		dev[32];

  pp = (PtProcessor)ProcessorList.SubList;
  *DeviceList = 0;
  while( pp != NULL )
  {
    cp = DeviceList;
    while( *cp )
    {
      cp1 = dev;
      while( ( *cp != 0 ) && ( *cp != ' ' ) )
		*( cp1++ ) = *( cp++ );
      *cp1 = 0;
      if( *cp )
		++cp;
      if( _fstrncmp( dev,pp->Device,33 ) == 0 )
      {
	Found = TRUE;
	break;
      }
    }
    if( !Found )
    {
      _fstrcat( DeviceList,pp->Device );
      strcat( DeviceList," " );
    }
    pp = (PtProcessor)pp->BackLink;
  }
}

// Resets the flag fields of the operator nodes
void ResetAllNodeFlags()
{
  PtNodeInfo	p;
  PtBlock	bp;

  p = (PtNodeInfo)(ActBlock->NodeList).SubList;
  while( p != NULL )
  {
    if( p->Type & OPTYPE_BLOCK )
    {
      bp = ActBlock;
      ActBlock = ActBlock->SearchBlockName( p->Operator );
      if( ActBlock == NULL )
      {
	wsprintf( work,"ResetAllNodeFlags : No block %s",
			(LPSTR)p->Operator );
	FatalError( work );
      }
      ResetAllNodeFlags();
      ActBlock = bp;
    }
    if( p->GraphPtr != NULL )
	(p->GraphPtr)->ResetFlags();
    p = (PtNodeInfo)p->BackLink;
  }
}

// Returns a pointer to the operator name of a given
// node
LPSTR GetNodeOp( NODETYPE node )
{
  PtNodeInfo	p;

  p = (PtNodeInfo)(ActBlock->NodeList).Search( node,OPTYPE_ALL );
  return (LPSTR)p->Operator;
}

// Computes the ASAP values recursively for a subtree of nodes
void ComputeNodeASAP( PtNodeObject np,unsigned long starttime )
{
  unsigned long	Asap = 0L;
  unsigned long ExecTime;
  PtNodeObject	np1;
  PtNodeObject	np2;
  PtNodeInfo	ip;

// Return if dead end
  if( np == NULL )
	return;
  np->Ready = TRUE;
  np->Aux = starttime;
  ip = ((PtNodeInfo)(np->InfoNode));
  if( np->IsReady() )
  {
    Asap = 0L;			// If input node
// ASAP(i) = max(ASAP(predecessors))
    np1 = np->BackNode;
    while( np1 != NULL )
    {
      if( np1->Aux > Asap )
		Asap = np1->Aux;
      np1 = np1->BackNode;
    }
    np1 = np;
    while( np1 != NULL )
    {
      if( np1->Aux > Asap )
		Asap = np1->Aux;
      np1 = np1->ForwNode;
    }
  }
  else
	return;
  ip->ASAP = Asap;
// Terminate if output probe or delay input
  if( ip->Type == OPTYPE_IDELAY )
		return;
  wsprintf( work,"(get-candidate %s '(%s) '%s%u)\n",
	      (LPSTR)ip->Operator,
	      (LPSTR)DeviceList,
	      (LPSTR)ActBlock->BlockName,
	      ip->NodeNum );
  _fstrcpy( work,DRCIExecuteCommand( CommWindow, work ) );
  XDDeleteReply( CommWindow );
  if( sscanf( work,"(%s %lu)",work1,&ExecTime ) < 2 )
  {
    ErrorMessage( work );
    return;
  }
  ip->ExecTime = ExecTime;
  ExecTime += Asap;		// To save some additions
  np1 = np->BackNode;
  while( np1 != NULL )
  {
    if( np1->OutputNum != NOT_VALID )
    {
      np1->Aux = Asap;
      np2 = (PtNodeObject)np1->SubList;
      while( np2 != NULL )
      {
	ComputeNodeASAP( np2,ExecTime );
	np2 = (PtNodeObject)np2->BackLink;
      }
    }
    np1 = np1->BackNode;
  }
  np1 = np;
  while( np1 != NULL )
  {
    if( np1->OutputNum != NOT_VALID )
    {
      np1->Aux = Asap;
      np2 = (PtNodeObject)np1->SubList;
      while( np2 != NULL )
      {
	ComputeNodeASAP( np2,ExecTime );
	np2 = (PtNodeObject)np2->BackLink;
      }
    }
    np1 = np1->ForwNode;
  }
}

// Computes the ASAP values for all the nodes.
void ComputeASAP()
{
  PtNodeInfo	p;

  p = (PtNodeInfo)(ActBlock->NodeList).SubList;
  while( p != NULL )
  {
    if( ( ( p->Type == OPTYPE_PROBE ) && ( p->Outputs == 1 ) ) ||
	( ( p->Type == OPTYPE_OP ) && ( p->GraphPtr != NULL ) &&
	  (p->GraphPtr)->IsReady() ) ||
	( p->Type == OPTYPE_ODELAY ) )
		ComputeNodeASAP( p->GraphPtr,0 );
    p = (PtNodeInfo)p->BackLink;
  }
}

// Computes the ALAP value for one node
void ComputeNodeALAP( PtNodeObject np )
{
  PtNodeObject	np1;
  PtNodeObject	np2;
  PtNodeInfo	ip;
  int		Type;
  unsigned long	ALAP = 0x7FFFFFFF;

  if( np == NULL )
		return;
  ip = (PtNodeInfo)(np->InfoNode);
// Return if input probe
  if( ( ( ip->Type == OPTYPE_PROBE ) && ( ip->Outputs == 1 ) ) ||
      ( ip->Type == OPTYPE_ODELAY ) )
		return;
// Walking the node chain to find output nodes
  np1 = np->BackNode;
  while( np1 != NULL )
  {
    if( np1->OutputNum != NOT_VALID )
    {
      np2 = (PtNodeObject)np1->SubList;
      while( np2 != NULL )
      {
	ip = (PtNodeInfo)(np2->InfoNode);
	if( ip->ALAP < ALAP )
		ALAP = ip->ALAP;
	np2 = (PtNodeObject)np2->BackLink;
      }
    }
    np1 = np1->BackNode;
  }
  np1 = np;
  while( np1 != NULL )
  {
    if( np1->OutputNum != NOT_VALID )
    {
      np2 = (PtNodeObject)np1->SubList;
      while( np2 != NULL )
      {
	ip = (PtNodeInfo)(np2->InfoNode);
	if( ip->ALAP < ALAP )
		ALAP = ip->ALAP;
	np2 = (PtNodeObject)np2->BackLink;
      }
    }
    np1 = np1->ForwNode;
  }

// ALAP is 0 if there has been at least one path connected
// to an output of the node which ALAP was not propagated
// through
  ip = (PtNodeInfo)(np->InfoNode);
  if( ALAP == 0L )
		return;
  ip->ALAP = ALAP - ip->ExecTime;

// Propagating recursively the ALAP time
  np1 = np->BackNode;
  while( np1 != NULL )
  {
    if( np1->InputNum != NOT_VALID )
	ComputeNodeALAP( np1->UpNode );
    np1 = np1->BackNode;
  }
  np1 = np->ForwNode;
  while( np1 != NULL )
  {
    if( np1->InputNum != NOT_VALID )
	ComputeNodeALAP( np1->UpNode );
    np1 = np1->ForwNode;
  }
}

// Computes the ALAP values for all the nodes
void ComputeALAP()
{
  PtNodeInfo	p;
  PtNodeObject	op;
  int		Type;

  p = (PtNodeInfo)(ActBlock->NodeList).SubList;
  while( p != NULL )
  {
    Type = p->Type;
    if( ( ( Type == OPTYPE_PROBE ) && ( p->Inputs == 1 ) ) ||
	( Type == OPTYPE_IDELAY ) )
    {
      p->ALAP = p->ASAP;
      op = p->GraphPtr;
      if( op != NULL )
	ComputeNodeALAP( op->UpNode );
    }
    p = (PtNodeInfo)p->BackLink;
  }
}

// Activates a node variable. Gets a pointer of the
// output node to which the variable will be assigned,
// pointer to the processor's descriptor on which the variable
// will be allocated ,the variable ID, the size of the variable.
// Returns NULL, if the variable couldn't be allocated else
// a pointer to the variable descriptor
PtVariable InsertVariable( PtNodeObject tree, PtProcessor ProcPtr,
	unsigned long ID,unsigned long Size )
{
  BOOL		Global;
  PtChunk	ProcDesc,Chk;
  PtBlock	BlockPtr;
  unsigned long	ChkAddr;
  PtNodeObject	p,np;
  PtPTrans	pt;
  PtVariable	Var,VarDesc;
  unsigned	Cons = 0;

  Global = ActBlock->IsRoot;
  p = (PtNodeObject)tree->SubList;
  while( p != NULL )
  {
    ++Cons;
    p = (PtNodeObject)p->BackLink;
  }
// If we are not in the root block, we have to check
// whether the variable is used outside of the block
// as well and increase its "consume number" accordingly.
   if( !ActBlock->IsRoot )
   {
     pt = (PtPTrans)(ActBlock->PTList).SubList;
     while( pt != NULL )
     {
       if( ( pt->Dir == PTRANS_OUTPUT ) &&
	   ( pt->PNodeNum == tree->NodeNum ) &&
	   ( pt->PIONum == tree->OutputNum ) )
       {
	 np = ActBlockNodeInfo->GraphPtr;
	 np = np->Search( NOT_VALID, pt->NodeNum );
	 p = (PtNodeObject)np->SubList;
	 while( p != NULL )
	 {
	   Global = TRUE; // Block local but consumed at global level
	   ++Cons;
	   p = (PtNodeObject)p->BackLink;
	 }
       }
       pt = (PtPTrans)pt->BackLink;
     }
   }


  if( Global )
	BlockPtr = &RootBlock;
  else
	BlockPtr = ActBlock;
  Var = new Variable( ProcPtr->NodeNum,ID );
  Var->Consumed = Cons;
  Var->Size = Size;
  Var->GraphPtr = (PtListObject)tree;
  Var->ProcPtr = (PtListObject)ProcPtr;
  Var->BlockPtr = (PtListObject)BlockPtr;
  if( ( ProcDesc = (PtChunk)(BlockPtr->ChunkList).Search
		( ProcPtr->NodeNum ) ) == NULL )
		return NULL;

  if( tree->Auto )	// If the compiler allocates space for the variable
  {
    if( ( ChkAddr = ProcDesc->Alloc( Size ) ) == CHK_NOMEM )
		return NULL;
    Chk = new Chunk( ProcPtr->NodeNum,ID,ChkAddr,Size );
    Var->AssocChunk = Chk;
    Var->Address = ChkAddr;
    ProcDesc->Insert( Chk );
  }
  else
// The statically allocated variables reserve space - as no chunk
// was allocated it must be handled separately (handled by
// Chunk::Insert else
	ProcDesc->Avail -= Size;

  if( ( VarDesc = (PtVariable)(BlockPtr->VariableList).Search( ProcPtr->NodeNum ) ) == NULL )
		return NULL;
  VarDesc->Insert( Var );
  tree->NodeVariable = Var;
  return Var;
}

// Creates an alias node and returns a pointer to it.
PtNodeInfo CreateAliasNode( PtNodeInfo ninf )
{
  PtNodeInfo	nninf = NULL;

  if( ninf->GraphPtr != NULL )
  {
    nninf = new NodeInfo( ninf->NodeNum,ninf->Type,
			ninf->Inputs,ninf->Outputs,
			ninf->Operator );
    nninf->GraphPtr = ninf->GraphPtr;
    nninf->ASAP = ninf->ASAP;
    nninf->ALAP = ninf->ALAP;
    nninf->SchedProc = ninf->SchedProc;
    nninf->Visited = ninf->Visited;
    nninf->Alias = ninf;
  }
  return nninf;
}

// Inserts the node identified by its node number into the
// ready list.
void InsertReadyList( PtNodeInfo ninf )
{
  PtNodeInfo	nninf;

  if( ( nninf = CreateAliasNode( ninf ) ) != NULL )
	ReadyList.Insert( nninf );
}

// Creates the ready node list. Takes the probes
// and inserts the lists whose all inputs are ready
// into the ready list.
void BuildReadyList()
{
  PtNodeInfo	p;
  PtNodeInfo	ninf;
  PtNodeObject	nptr;
  PtNodeObject	mptr;
  unsigned long size;

  p = (PtNodeInfo)(ActBlock->NodeList).SubList;
  while( p != NULL )
  {
// If probe
    if( ( p->Type == OPTYPE_PROBE ) && ( p->GraphPtr != NULL ) )
    {
      nptr = (PtNodeObject)(((PtNodeObject)p->GraphPtr)->SubList);
      if( ( nptr != NULL ) && ( p->Outputs == 1 ) )
		InsertReadyList( p );
    }
    else
// This will enable
// - operator with no inputs
// - operator made ready in an embedded block by dataflow at upper level
    if(	( p->Type == OPTYPE_OP ) && (p->GraphPtr)->IsReady() )
		InsertReadyList( p );
// if delay output
    else
    if( p->Type == OPTYPE_ODELAY )
		InsertReadyList( p );
    p = (PtNodeInfo)p->BackLink;
  }
}

// Makes the subchain tasks of the given node active.
// Inserts the subchain nodes into the ReadyList and
// sets their Ready flag.
void MakeSubchainReady( PtNodeObject Object )
{
  PtNodeObject	p;
  PtNodeInfo	ninf;
  unsigned long	FTime;

  if( Object->OutputNum != NOT_VALID )	// If it is an output node
  {
    FTime = ((PtNodeInfo)(Object->InfoNode))->FinishTime;
    p = (PtNodeObject)Object->SubList;
    while( p != NULL )
    {
      p->Ready = TRUE;		// It is ready now
      p->VarReadyTime = FTime;
      if( p->IsReady() )	// If node is ready
      {
	ninf = (PtNodeInfo)(p->InfoNode);
	ninf->CalculateStartTime();
	InsertReadyList( ninf );
      }
      p = (PtNodeObject)p->BackLink;
    }
  }
}

// Activates all the operator inputs connected to all the
// outputs of the given node
void MakeSubchainsReady( PtNodeObject Object )
{
  PtNodeObject	p;
  PtNodeInfo	ninf;

  ninf = (PtNodeInfo)Object->InfoNode;
// If the executed node was a delay output, its associated
// input may have got ready. The condition must be checked
// now else the delay input may never be scheduled if the
// node connected to the input of the delay has already
// been executed
  if( ninf->Type == OPTYPE_ODELAY )
  {
    ninf = (PtNodeInfo)(ActBlock->NodeList).Search( ninf->NodeNum,OPTYPE_IN );
    if( ( ninf != NULL ) && !ninf->Executed &&
	(ninf->GraphPtr)->IsReady() )
		InsertReadyList( ninf );
  }
  p = Object->ForwNode;
  while( p != NULL )
  {
    MakeSubchainReady( p );
    p = p->ForwNode;
  }
  p = Object;
  while( p != NULL )
  {
    MakeSubchainReady( p );
    p = p->BackNode;
  }
}

// Kills the variable associated with the given input node
// if the time-to-live was exceeded
void KillInputVar( PtNodeObject Object )
{
  PtNodeObject	Parent;
  PtVariable	Var,OV;
  PtBlock	BlockPtr;
  PtChunk	ProcList;
  PtChunk	pchk;
  int		ProcNum;

  if( Object->InputNum != NOT_VALID )	// if it is an input node
  {
// For block input nodes which do not have parent but
// for which node variable has been propagated from
// the block inputs
    if( ( Parent = Object->UpNode ) == NULL )
		Parent = Object;
    if( ( Var = Parent->NodeVariable ) == NULL )
		return;
    if( !( --(Var->Consumed) ) )	// Removing variable if finally consumed
    {
// Walks the variable chain to kill all the instances of the
// variable allocated on different processors
      OV = Var->VarLink;
      do
      {
	ProcNum = Var->NodeNum;
	BlockPtr = (PtBlock)Var->BlockPtr;
	ProcList = (PtChunk)(BlockPtr->ChunkList).Search( ProcNum );
	pchk = Var->AssocChunk;
	if( pchk != NULL )	// Assure that it was not auto variable
	{
	  ProcList->Remove( pchk );
	  delete pchk;
	}
	Var = Var->VarLink;
      }
      while( Var != OV );
    }
  }
}

// Kills the variables associated with the inputs of
// the given node. Gets a pointer to one element of
// the node link.
void KillInputVars( PtNodeObject Object )
{
  PtNodeObject	p;

  p = Object->ForwNode;
  while( p != NULL )
  {
    KillInputVar( p );
    p = p->ForwNode;
  }
  p = Object;
  while( p != NULL )
  {
    KillInputVar( p );
    p = p->BackNode;
  }
}

// Chooses the correct instance of the variable for the
// given processor. Gets a pointer to one instance of the
// variable and a pointer to the processor descriptor and
// returns a pointer to the correct instance or NULL if
// the variable does not exist on the given processor
PtVariable ChooseInstance( PtVariable AVar, PtProcessor Proc )
{
  PtVariable	p;

  p = AVar;
  do
    if( (PtProcessor)p->ProcPtr == Proc )
	return p;
    else
	p = p->VarLink;
  while( p != AVar );
  return NULL;
}

// Asks the number of the communication channel associated with
// a given processor pair. Searches the cache and if there is
// no hit, asks the XLisp hardware database. Gets the processor
// pair, returns the channel number.
int AskChannelNum( int FromProc, int ToProc )
{
  int	i;

  i = ( CCachePtr == 0 ? CCACHESIZE-1 : CCachePtr-1 );
  for( ; i != CCachePtr ; i = ( i == 0 ? CCACHESIZE - 1 : i-1 ) )
    if( ( ccache[i].FromProc == FromProc ) && ( ccache[i].ToProc == ToProc ) )
		return ccache[i].ChannelNum;
  wsprintf( work,"(get-channel %u %u)\n", FromProc, ToProc );
  _fstrcpy( work,DRCIExecuteCommand( CommWindow, work ) );
  XDDeleteReply( CommWindow );
  if( sscanf( work,"%u",&i ) < 1 )
  {
    sprintf( work1,"get-channel %u %u returned : %s",FromProc,ToProc,work );
    Error2Message( work1 );
    FatalError();
  }
  ccache[CCachePtr].FromProc = FromProc;
  ccache[CCachePtr].ToProc = ToProc;
  ccache[CCachePtr].ChannelNum = i;
  CCachePtr = (++CCachePtr) % CCACHESIZE;
  return i;
}

// Inserts a communication activity pair.
// Gets the sender and the receiver processor the variable
// to be transferred,
// and the earliest scheduling time for the operation
// (used only in the case of receive operation)
// Returns a pointer to the created comm activity
void InsertCommActivity( PtProcessor SProc, PtProcessor RProc,
		    PtVariable Var,PtNodeObject Object )
{
  PtActivity	SNAct1,RNAct1,SAct,RAct,p;
  PtVariable	NVar;
  unsigned long	l;
  unsigned long	SSTime,SExecTime,SSurviveTime;
  unsigned long RSTime,RExecTime,RSurviveTime;
  unsigned sproc,rproc,ChannelNum;
  int		SSync,RSync;
  PtChannel CP,NCP;

  sproc = SProc->NodeNum;
  rproc = RProc->NodeNum;

// Activity top level descriptors

  SAct = (PtActivity)(ActBlock->ActivityList).Search( SProc->NodeNum );
  RAct = (PtActivity)(ActBlock->ActivityList).Search( RProc->NodeNum );

// Filling channel descriptor
  ChannelNum = AskChannelNum( rproc,sproc );
  if( ( CP = (ActBlock->Channels).Search( ChannelNum ) ) == NULL )
  {
    sprintf( work1,"Reference to not existing channel number : %u",ChannelNum );
    Error2Message( work1 );
    FatalError();
  }
// Descriptor of our activity
  NCP = new Channel( ChannelNum );
  CP->Insert( NCP );

// Sender block

  wsprintf( work,"(get-commcost %u %u 'S %lu)\n",
			sproc,rproc,Var->Size );
  _fstrcpy( work,DRCIExecuteCommand( CommWindow, work ) );
  XDDeleteReply( CommWindow );
  if( sscanf( work,"(%lu %lu %u)",&SExecTime,&SSurviveTime,&SSync ) < 3 )
  {
    sprintf( work1,"get-commcost %u %u 'S %lu returned : %s",
		sproc,rproc,Var->Size,work );
    Error2Message( work1 );
    FatalError();
  }

  SSTime = SAct->STime;
  SNAct1 = new Activity( 0,ACT_SEND,SSTime,rproc );
  SNAct1->ExecTime = SExecTime;
  SNAct1->VarField = Var;

// Modifying channel descriptor and finding out schedule time
  SSTime = maxl( CP->SendEndOccupied, SSTime );
  NCP->SendStartOccupied = CP->SendStartOccupied = SSTime;
  NCP->SendEndOccupied = CP->SendEndOccupied = SSTime + SExecTime + SSurviveTime;

  SNAct1->STime = SSTime;
  SAct->Insert( SNAct1 );
  SAct->STime = SSTime + SExecTime;

// New var on the receiving processor
  NVar = InsertVariable( Object,RProc,Var->ID,Var->Size );
  if( NVar == NULL )
  {
    sprintf( work1,"Couldn't create variable on proc #%u of size %lu",
		rproc,Var->Size );
    Error2Message( work1 );
    FatalError();
  }

// Links the new var into the circular list of the instances
// of the same variable on different processors
  NVar->VarLink = Var->VarLink;
  Var->VarLink = NVar;

// Receive block

  wsprintf( work,"(get-commcost %u %u 'R %lu)\n",
			sproc,rproc,Var->Size );
  _fstrcpy( work,DRCIExecuteCommand( CommWindow, work ) );
  XDDeleteReply( CommWindow );
  if( sscanf( work,"(%lu %lu %u)",&RExecTime,&RSurviveTime,&RSync ) < 2 )
  {
    sprintf( work1,"get-commcost %u %u 'R %lu returned : %s",
		sproc,rproc,NVar->Size,work );
    Error2Message( work1 );
    FatalError();
  }

  RNAct1 = new Activity( 0,ACT_RECEIVE,0L,sproc );
  RNAct1->ExecTime = RExecTime;
  RNAct1->VarField = NVar;

  RSTime = maxl( RAct->STime,SSTime );

// Which is latest ? the earliest receive schedule time,
// the end of the channel occupation for receive or the earliest
// time when the processor is free ? We will schedule our
// receive activity then
  RSTime = maxl( CP->RecEndOccupied, RSTime );

// The channel is occupied for the whole period of the
// transaction.
  NCP->RecStartOccupied = CP->RecStartOccupied = RSTime;	// has no meaning now
  RAct->CommInProgress = NCP->RecEndOccupied = CP->RecEndOccupied
		       = RSTime + RExecTime + RSurviveTime;

  RNAct1->STime = RSTime;
  RAct->Insert( RNAct1 );
  RAct->STime = RSTime + RExecTime;

// If synchronization is necessary the send operation must be
// timed to the receive operation
  if( SSync )
  {
    NCP->SendStartOccupied = CP->SendStartOccupied = CP->RecStartOccupied;
    NCP->SendEndOccupied = CP->SendEndOccupied = CP->RecEndOccupied;
    SNAct1->STime = RSTime;
    SAct->STime = RSTime + SExecTime;
  }

  RNAct1->PartnerAct = SNAct1;
  SNAct1->PartnerAct = RNAct1;
}


// Fetches all the input variables of the node. Does nothing
// if there are instances of all the input variables on the
// processor on which the node will be executed, inserts
// communication activities else. Gets a pointer to the
// processor descriptor and to the node object
int FetchInputVar( PtProcessor ProcPtr, PtNodeObject Object )
{
  PtVariable	Var;

  if( Object->InputNum != NOT_VALID )	// If input
  {
// It can happen that a node has inputs while no prede-
// cessors if it in an embedded block and receives its
// input from an embedded block input. In this case
// Propagate* routines must already have assigned
// the node variable
    if( Object->UpNode == NULL )
	Var = Object->NodeVariable;
    else
	Var = (Object->UpNode)->NodeVariable;
    if( Var == NULL )
    {
      wsprintf( work1,"FetchInputVar : var not allocated %u_(%u,%u)",
			Object->NodeNum,Object->InputNum,
			Object->OutputNum );
      Error2Message( work1 );
      FatalError();
    }
// Finds out whether comm activity is needed
    if( ChooseInstance( Var, ProcPtr ) == NULL )
      InsertCommActivity( (PtProcessor)Var->ProcPtr,ProcPtr, Var,Object );
  }
  return 0;
}

// Fetches the input variables of a node from other processors
// if necessary. Gets a pointer to the processor descriptor and
// the node info pointer
int FetchInputVars( PtProcessor ProcPtr, PtNodeInfo NNode )
{
  PtNodeObject	Object,p;

  Object = NNode->GraphPtr;
  p = Object->ForwNode;
  while( p != NULL )
  {
    if( FetchInputVar( ProcPtr,p ) )
	return 1;
    p = p->ForwNode;
  }
  p = Object;
  while( p != NULL )
  {
    if( FetchInputVar( ProcPtr,p ) )
	return 1;
    p = p->BackNode;
  }
  return 0;
}

// Creates the variable belonging to the output node.
// Gets a pointer to the node element and the processor number.
// Returns TRUE if the variable couldn't be created.
BOOL CreateOutputVar( PtProcessor ProcPtr, PtNodeObject Object )
{
  if( Object->OutputNum != NOT_VALID )	// if output node
  {
    if( InsertVariable( Object,ProcPtr,VarID++,Object->Size ) == NULL )
			return TRUE;
    Object->Ready = TRUE;		// Variable created
  }
  return FALSE;
}

// Create the variables belonging to the node. Gets the processor
// ID and a pointer to one node element. Returns 1 if error
int CreateOutputVars( PtProcessor ProcPtr,PtNodeObject Object )
{
  PtNodeObject	p;

  p = Object->ForwNode;
  while( p != NULL )
  {
    CreateOutputVar( ProcPtr,p );
    p = p->ForwNode;
  }
  p = Object;
  while( p != NULL )
  {
    CreateOutputVar( ProcPtr,p );
    p = p->BackNode;
  }
  return 0;
}

// Kills a variable associated with the given output node
// if no other operators consumes that
void KillUnusedOutputVar( PtNodeObject Object )
{
  PtVariable	Var;
  PtBlock	BlockPtr;
  PtChunk	ProcList;
  PtChunk	pchk;
  int		ProcNum;

  if( Object->OutputNum != NOT_VALID )	// if it is an output node
  {
    if( ( Var = Object->NodeVariable ) == NULL )
		return;
    if( Var->Consumed == 0 )	// if nothing consumes this
    {
      ProcNum = Var->NodeNum;
      BlockPtr = (PtBlock)Var->BlockPtr;
      ProcList = (PtChunk)(BlockPtr->ChunkList).
		 Search( ProcNum );
      pchk = Var->AssocChunk;
      ProcList->Remove( pchk );
      delete pchk;
    }
  }
}

// Kills the variables belonging to unconnected outputs
int KillUnusedOutputVars( PtNodeObject Object )
{
  PtNodeObject	p;

  p = Object->ForwNode;
  while( p != NULL )
  {
    KillUnusedOutputVar( p );
    p = p->ForwNode;
  }
  p = Object;
  while( p != NULL )
  {
    KillUnusedOutputVar( p );
    p = p->BackNode;
  }
  return 0;
}

// Finds out whether the IPF has changed and inserts IPF
// change activity if yes. Gets the
void CheckIPFChange( PtProcessor Proc )
{
  PtActivity	PAct,NAct;

  if( ( Proc->UpScale != UpScale ) ||
      ( Proc->DownScale != DownScale ) )
  {
    PAct = (PtActivity)(ActBlock->ActivityList).Search( Proc->NodeNum );
    NAct = new Activity( 0,ACT_SAMPLERATE,PAct->STime,0L );
    NAct->UpScale = UpScale;
    NAct->DownScale = DownScale;
    PAct->Insert( NAct );
    Proc->UpScale = UpScale;
    Proc->DownScale = DownScale;
  }
}

// Inserts an activity into the activity list of the given
// processor. Gets the processor ID and a pointer to the
// node's nodeinfo.
void InsertActivity( PtProcessor PProc, PtNodeInfo NInfo )
{
  PtActivity	PAct,NAct;
  PtNodeInfo	OInfo;
  unsigned long	exectime,utmp;
  int		ActID;

  PAct = (PtActivity)(ActBlock->ActivityList).Search( PProc->NodeNum );
  switch( NInfo->Type )
  {
    case OPTYPE_OP:
	ActID = ACT_OPERATOR;
	break;
    case OPTYPE_ODELAY:
	ActID = ACT_GETDELAY;
	break;
    case OPTYPE_IDELAY:
	ActID = ACT_PUTDELAY;
	break;
    case OPTYPE_PROBE:
	ActID = NInfo->Inputs == NOT_VALID ?
			 ACT_OUTPROBE : ACT_INPPROBE;
	break;
  }
  OInfo = NInfo->Alias;
// Scheduling time is modified if the operation has to wait
// for a variable whose communication is made in parallel
  if( PAct->CommInProgress > PAct->STime )
		PAct->STime = PAct->CommInProgress;

  if( PAct->STime < OInfo->StartTime )
	PAct->STime = OInfo->StartTime;
  else
	OInfo->StartTime = PAct->STime;
  NAct = new Activity( NInfo->NodeNum, ActID, PAct->STime,0 );
  NAct->UpScale = UpScale;
  NAct->DownScale = DownScale;
  utmp = ( UpScale * LDownScale )/( DownScale * LUpScale );
  PAct->Insert( NAct );

  wsprintf( work,"(get-exectime %s '%s '%s%u)\n",(LPSTR)NInfo->Operator,
	      (LPSTR)PProc->Device,(LPSTR)ActBlock->BlockName,
	      NInfo->NodeNum );
  _fstrcpy( work,DRCIExecuteCommand( CommWindow, work ) );
  XDDeleteReply( CommWindow );
  if( sscanf( work,"%lu",&exectime ) < 1 )
  {
    wsprintf( work1,"get-exectime %u '%s '%s '%s%u returned : %s\n",
	      (LPSTR)PProc->Device,(LPSTR)ActBlock->BlockName,
	      NInfo->NodeNum,(LPSTR)work );
    Error2Message( work1 );
    FatalError();
  }
// Exectime is multiplied by the loop count by which the operator
// is repeated if it is executed at higher sample rate
  exectime *= utmp;
  NAct->ExecTime = exectime;
  PAct->STime += exectime;
  PProc->OccupiedTime = PAct->STime;
  OInfo->FinishTime = PAct->STime;
  OInfo->Executed = TRUE;
}

// Launches the execution of a node. The node is removed from
// the ready list, its output variables are allocated and
// the task is placed into the activity queue of the given
// processor.Returns 0 if the node was executed succesfully else
// 1, if out of memory
int BeginExecuteNode( PtProcessor ProcPtr , PtNodeInfo NNode )
{
  PtNodeObject	NObject;
  PtNodeInfo	ONode;

  ONode = NNode->Alias;
  NObject = NNode->GraphPtr;
  ReadyList.Remove( (PtListObject)NNode );

// Inserts ACT_SAMPLERATE activity if necessary
  CheckIPFChange( ProcPtr );

// Fetching input variables from other processors if necessary
// (avoids considering the fake input nodes of delay outputs)
  if( NNode->Type != OPTYPE_ODELAY )
	if( FetchInputVars( ProcPtr,NNode ) )
			return 1;

// Delay inputs have fake output nodes which belong to the
// delay output section. For this reason creating these
// output variables must be avoided. The same is true for
// output probes where an output node is created so that IPFs
// can be fetched for those probes.
  if( ( NNode->Type != OPTYPE_IDELAY )
      && !( ( NNode->Type == OPTYPE_PROBE ) &&
	    ( NNode->Inputs == 1 ) ) )
    if( CreateOutputVars( ProcPtr,NObject ) )
			return 1;

  InsertActivity( ProcPtr,NNode );

  ONode->SchedProc = (PtListObject)ProcPtr;

  return 0;
}

// Finishes node execution. The input variables of the node
// are freed if their time-to-live exceeded, the operators
// connected to its output nodes are activated.
void FinishExecuteNode( PtProcessor ProcPtr,PtNodeInfo NNode )
{
  PtNodeInfo	ONode;
  PtNodeObject	NObject;

  ONode = NNode->Alias;
  NObject = NNode->GraphPtr;

// A delay output has fake input node (which are connected
// to the delay input). Killing those input variables
// must be avoided.
  if( NNode->Type != OPTYPE_ODELAY )
	KillInputVars( NObject );

// See note at CreateOutputVars, some lines above
  if( NNode->Type != OPTYPE_IDELAY )
  {
    MakeSubchainsReady( NObject );
    KillUnusedOutputVars( NObject );
  }

  ONode->Executed = TRUE;
  delete NNode;		// As it was only a copy of the NodeList
			// element
}

// Reads token. Gets a pointer to the character pointer
// in the stream and reads the string until a space, a 0
// character or a right parenthesis is found.
void ReadToken( char *cp[] , char token[] )
{
  while( ( **cp != ' ' ) && ( **cp != 0 ) && ( **cp != ')' ) )
	*( token++ ) = *( (*cp)++ );
  *token = 0;
}

// Asks XLisp operator database about the processors
// available for a given operator and about the execution
// times and builds the device list. Gets the nodeinfo
// structure of the node to be examined.
void BuildDeviceListForOperator( PtNodeInfo NI )
{
  char		D[32];
  unsigned long	ET;
  PtExecTimeList	p;
  int		Items = 0;
  char		*cp;

  wsprintf( work,"(get-exectimes %s '%s%u)\n",
	    (LPSTR)NI->Operator,
	    (LPSTR)ActBlock->BlockName,
	    NI->NodeNum );
  _fstrcpy( work,DRCIExecuteCommand( CommWindow, work ) );
  XDDeleteReply( CommWindow );
  cp = work+1;	// Skipping starting (
  while( ( *cp != 0 ) && ( *cp != ')' ) )
  {
    D[0] = 0;
    ReadToken( &cp,D );
    if( *D )
    {
      ++cp;		// Skipping space
      ReadToken( &cp,work1 );
      if( sscanf( work1,"%lu",&ET ) < 1 )
      {
	wsprintf( work1,"get-exectimes %s '%s%u : %s",
		(LPSTR)NI->Operator,
		(LPSTR)ActBlock->BlockName,
		NI->NodeNum,
		(LPSTR)work );
	Error2Message( work1 );
	FatalError();
      }
      if( *cp == ' ' )
		++cp;
      p = new ExecTimeList( D,ET );
      TimeList.Insert( p );
      ++Items;
    }			// if( *D )
  }			// while

  if( !Items )
  {
    sprintf( work,"No processor can execute node %u",NI->NodeNum );
    Error2Message( work );
    FatalError();
  }
}

// Estimates the communication cost between two nodes.
// Gets the processor descriptor on which the node
// is supposed to execute, and the info ptr of a node
// is to be placed. Returns the estimated comm. time.
unsigned long EstimateNodeCommTime( PtProcessor ProcPtr, PtNodeObject Object )
{
  PtVariable	Var;
  int		pairproc;
  unsigned long	est;

  if( Object->InputNum != NOT_VALID )
  {
// See note at FetchInputVariable
    if( Object->UpNode == NULL )
	Var = Object->NodeVariable;
    else
	Var = (Object->UpNode)->NodeVariable;
    if( Var == NULL )
    {
      wsprintf( work1,"EstimateNodeCommTime : var not allocated %u_(%u,%u)",
			Object->NodeNum,Object->InputNum,
			Object->OutputNum );
      Error2Message( work1 );
      FatalError();
    }

// Finds out whether comm activity is needed
    if( ChooseInstance( Var, ProcPtr ) == NULL )
    {
      pairproc = ((PtProcessor)(Var->ProcPtr))->NodeNum;
      if( ProcPtr->NodeNum == pairproc )
		return 0L;
      else
      {
	wsprintf( work,"(get-commcost-est %u %u %lu)\n",
		  pairproc,ProcPtr->NodeNum,Var->Size );
	_fstrcpy( work,DRCIExecuteCommand( CommWindow, work ) );
	XDDeleteReply( CommWindow );
	if( sscanf( work,"%lu",&est ) < 1 )
	{
	  sprintf( work1,"get-commcost-est %u %u %lu : %s",
		  pairproc,ProcPtr->NodeNum,Var->Size,work );
	  Error2Message( work1 );
	  FatalError();
	}	// if( sscanf ...
	return est;
      }		// if( ProcPtr->NodeNum ...
    }		// if( ChooseInstance ...
  }		// if( Object ...
  return 0L;
}

// Computes approximate communication time of fetching the
// input variables if a given node is scheduled on a given
// processor. Gets a pointer to the nodeinfo structure,
// a pointer to the processor descriptor, returns the
// estimated communication time.
unsigned long EstimateCommTime( PtProcessor pp, PtNodeInfo ip )
{
  PtNodeObject	np;
  unsigned long	est = 0L;

  np = ip->GraphPtr;
  np = np->ForwNode;
  while( np != NULL )
  {
    est += EstimateNodeCommTime( pp,np );
    np = np->ForwNode;
  }
  np = ip->GraphPtr;
  while( np != NULL )
  {
    est += EstimateNodeCommTime( pp,np );
    np = np->BackNode;
  }
  return est;
}

// Returns the earliest time at which the node can be scheduled.
// Checks the input nodes of the operator and finds the maximum
// VarReadyTime. Gets the nodeinfo ptr, returns the earliest
// scheduling time.
unsigned long EarliestSchedTime( PtNodeInfo ip )
{
  PtNodeObject	np;
  unsigned long maxt = 0L;

  np = (ip->GraphPtr)->ForwNode;
  while( np != NULL )
  {
    if( ( np->InputNum != NOT_VALID ) && ( np->VarReadyTime > maxt ) )
	maxt = np->VarReadyTime;
    np = np->ForwNode;
  }
  np = ip->GraphPtr;
  while( np != NULL )
  {
    if( ( np->InputNum != NOT_VALID ) && ( np->VarReadyTime > maxt ) )
	maxt = np->VarReadyTime;
    np = np->ForwNode;
  }
  return maxt;
}

// Searches for the candidate processor for the given node
// Gets a pointer to the NodeInfo structure, returns a pointer
// to the processor chosen
PtProcessor SearchACandidate( PtNodeInfo np )
{
  unsigned long	s,et,smin,cest = 0L;
  PtExecTimeList     etp;
  PtProcessor	sp = NULL,pp;
  PtNodeInfo	op;

  if( np->Type == OPTYPE_PROBE )
// Forcing probes to their initial processor
	sp = (PtProcessor)np->SchedProc;
  else
// Delay inputs must be scheduled on the same processor with
// their corresponding delay outputs
  if( np->Type == OPTYPE_IDELAY )
  {
    op = (ActBlock->NodeList).Search( np->NodeNum,OPTYPE_ODELAY );
    return (PtProcessor)op->SchedProc;
  }
  else
  {
    BuildDeviceListForOperator( np );
// Checking finishing time on each processor supported by the
// node
    pp = (PtProcessor)ProcessorList.SubList;
    while( pp != NULL )
    {
// et : execution time on that processor
      if( ( etp = TimeList.SearchDevice( pp->Device ) ) != NULL )
      {
	et = etp->ExecTime;
// Fake input nodes of delay outputs ... Same old story
	if( np->Type != OPTYPE_ODELAY )
		cest = EstimateCommTime( pp,np );
	s = EarliestSchedTime( np ) + cest + et;
	if( ( sp == NULL ) || ( s < smin ) )
	{
	  sp = pp;
	  smin = s;
	}
      }
      pp = (PtProcessor)pp->BackLink;
    }
    TimeList.ClearSubList();
  }
  return sp;
}

// Updates the virtual time. The virtual time is maintained for
// tracking the position in the ASAP-ALAP schedule graphs.
// The virtual time is the smallest ALAP time among the
// ready nodes.
void UpdateVirtualTime()
{
  PtNodeInfo	p;
  unsigned long	vt = 0xFFFFFFFF;

  p = (PtNodeInfo)ReadyList.SubList;
  while( p != NULL )
  {
    if( p->ALAP < vt )
	vt = p->ALAP;
    p = (PtNodeInfo)p->BackLink;
  }
  VirtualTime = vt;
}

// Calculates the actual weight for the given node entry.
// The weight is the variable size necessary for the node.
// It is negative if the input node entry consumes this
// variable and positive if the output node entry produces
// the variable
long CalculateWeight( PtNodeObject Object, int Type )
{
  if( ( Object->InputNum != NOT_VALID ) &&
      ( Type != OPTYPE_ODELAY ) )	// if input
	return -(Object->Size);
  else					// if output
  if( Object->Auto )			// No memory requirement
					// if not auto variable
	return Object->Size;
  else
	return 0;
}

// Search candidate node in the ready list.
PtNodeInfo SearchCandidateNode()
{
  PtNodeInfo	p, mp = NULL;
  PtNodeObject	op;
  float		mrw,aw,fw,minweight;
  long		absweight,weight,nw;
  BOOL		minreg = FALSE;
  unsigned	PUS,PDS;

  p = (PtNodeInfo)ReadyList.SubList;	// Traverse ready list for
  minweight = 0;
  while( p != NULL )			// minimum weight
  {
    weight = 0;
    absweight = 0L;
    if( !p->Visited )
    {
      wsprintf( work1,"Node %u was not visited by IPF checker",
		     p->NodeNum );
      FatalError( work1 );
    }
    p->GetScales( &PUS,&PDS );
    op = (p->GraphPtr)->ForwNode;
    if( ( UpScale == PUS ) && ( DownScale == PDS ) )
    {
      while( op != NULL )
      {
	nw = CalculateWeight( op,p->Type );
	weight += nw;
	if( nw < 0 )
	  absweight -= nw;
	else
	  absweight += nw;
	op = op->ForwNode;
      }
      op = p->GraphPtr;
      while( op != NULL )
      {
	nw = CalculateWeight( op,p->Type );
	weight += nw;
	if( nw < 0 )
	  absweight -= nw;
	else
	  absweight += nw;
	op = op->BackNode;
      }
      if( absweight == 0L )
	mrw = 0;
      else
	mrw = (float)weight/(float)absweight;
      aw = (float)( p->ALAP - VirtualTime );
      fw = MemReqWeight * mrw + FreedomWeight * aw;
      if( ( fw < minweight ) || !minreg )
      {
	minweight = fw;
	minreg = TRUE;
	mp = p;
      }
    }
    p = (PtNodeInfo)p->BackLink;
  }		// while( p != NULL )
  return mp;
}

// Updates UpScale and DownScale values - looks for the IPF
// with the maximum number of nodes in the ready list.
void UpdateIPF()
{
  PtNodeInfo	p,p1;
  PtActivity	ap;
  unsigned long	max = 0L;
  unsigned	PUS,PDS,US1,DS1;

  p = (PtNodeInfo)ReadyList.SubList;
  while( p != NULL )
  {
    p->Aux = 0L;
    p = (PtNodeInfo)p->BackLink;
  }
  p = (PtNodeInfo)ReadyList.SubList;
  while( p != NULL )
  {
    if( p->Aux == 0L )
    {
      ++(p->Aux);
      p1 = (PtNodeInfo)p->BackLink;
      p->GetScales( &PUS, &PDS );
      while( p1 != NULL )
      {
	if( !p1->GetScales( &US1,&DS1 ) )
	{
	  wsprintf( work,"IPF not set correctly for node %u in block %s",
	  p1->NodeNum,(LPSTR)ActBlock->BlockName );
	  FatalError( work );
	}
	if( ( US1 == PUS ) && ( DS1 == PDS ) )
	{
	  p1->Aux = 1L;		// Can be 0<Aux<2 :-)
	  ++(p->Aux);
	}
	p1 = (PtNodeInfo)p1->BackLink;
      }
    }
    p = (PtNodeInfo)p->BackLink;
  }
  p = (PtNodeInfo)ReadyList.SubList;
  while( p != NULL )
  {
    if( p->Aux > max )
    {
      p->GetScales( &UpScale,&DownScale );
      max = p->Aux;
    }
    p = (PtNodeInfo)p->BackLink;
  }
}

// Makes translation inputs ready. Scans the translation
// input list and sets the ready flag of their inputs.
void MakeTransInputsReady()
{
  PtPTrans	pt;
  PtNodeInfo	ip;
  PtNodeObject  np;
  NODETYPE	nn,in;

  pt = (PtPTrans)(ActBlock->PTList).SubList;
  while( pt != NULL )
  {
    if( ( pt->Dir == PTRANS_INPUT ) &&
	( pt->NodeNum != 0 ) ) // if not control input
    {
      nn = pt->PNodeNum;
      in = pt->PIONum;
      ip = (PtNodeInfo)(ActBlock->NodeList).Search( nn,OPTYPE_ALL );
      if( ip == NULL )
      {
	wsprintf( work1,"MakeTransInputReady : Node %u not found",
			nn );
	FatalError( work1 );
      }
      np = (PtNodeObject)(ip->GraphPtr);
      np = np->Search( in,NOT_VALID );
      if( np == NULL )
      {
	wsprintf( work,"MakeTransInputReady : Input %u of node %u not found",
			  in,nn );
	FatalError( work );
      }
      np->Ready = TRUE;
    }
    pt = (PtPTrans)pt->BackLink;
  }
}

// 		        Scheduler routine
//		        *** TEMPORARY ***
//		(Remains temporary forever :-) )
// This routine does the whole scheduling task. The output of
// this routine is the activity list for the target processors
// Returns 0 if OK else error code
// 1, if out of memory
// 10, if no node could be inserted into the ready list
int Schedule()
{
  PtNodeInfo	p,mp;
  PtProcessor	pp;
  int		ecode,i;

  for( i = 0 ; i < CCACHESIZE ; i++ )
  {
    ccache[i].FromProc = 0;
    ccache[i].ToProc = 0;
  }
  CCachePtr = 0;
  VirtualTime = 0L;
  MemReqWeight = 0.1;
  FreedomWeight = 0.9;
  CreateDeviceList();
  ComputeASAP();
  ComputeALAP();
  ResetAllNodeFlags();
  MakeTransInputsReady();
  BuildReadyList();
  UpdateIPF();
  if( ReadyList.SubList == NULL )
		return 1;
  while( ReadyList.SubList != NULL )
  {
    while( ReadyList.SubList != NULL )
    {
      mp = SearchCandidateNode();
      if( mp == NULL )
      {
	UpdateIPF();
	mp = SearchCandidateNode();
      }
// mp points to the candidate node
      if( mp != NULL )
      {
	pp = SearchACandidate( mp );
	if( pp == NULL )
	{
	  sprintf( work,"No processor can execute node %u",mp->NodeNum );
	  Error2Message( work );
	  FatalError();
	}
	ecode = BeginExecuteNode( pp,mp );
	if( ecode ) return ecode;
	FinishExecuteNode( pp,mp );
      }
      UpdateVirtualTime();
    }		// while( ReadyList.SubList != NULL )
  }		// while( ReadyList.SubList != NULL )
  return 0;
}


// Checks the scheduling created by Schedule() and sends
// warning messages if anomalies have been found
// Checks
// - if all the nodes have been executed,
// - if all output variables have been created
void ScheduleCheck()
{
  PtNodeInfo	p;
  PtNodeObject	np;
  PtNodeObject	Parent;

  p = (PtNodeInfo)(ActBlock->NodeList).SubList;
  while( p != NULL )
  {
    if( !p->Executed )
    {
      switch( p->Type )
      {
	case OPTYPE_OP:
		strcpy( work1,"Operator" );
		break;
	case OPTYPE_IDELAY:
		strcpy( work1,"Delay input" );
		break;
	case OPTYPE_ODELAY:
		strcpy( work1,"Delay output" );
		break;
	case OPTYPE_PROBE:
		strcpy( work1,"Probe" );
		break;
	case OPTYPE_BLOCK:
		strcpy( work1,"Block" );
		break;
      }
      sprintf( work2," %u was not executed",p->NodeNum );
      strcat( work1,work2 );
      WarningMessage( work1 );
    }
    if( ( p->Type == OPTYPE_PROBE ) && ( p->Inputs == 1 ) )
    {
      if( ( np = p->GraphPtr ) == NULL )
      {
	sprintf( work2,"Output probe %u is not connected",p->NodeNum );
	WarningMessage( work2 );
      }
      else
      {
	if( ( ( Parent = np->UpNode ) != NULL ) &&
	    ( Parent->NodeVariable == NULL ) )
	{
	  sprintf( work2,"Output variable %u was not generated",p->NodeNum );
	  WarningMessage( work2 );
	}
      }
    }
    p = (PtNodeInfo)p->BackLink;
  }
}

// Propagate one input variable from the source to the
// embedded operator's input
void PropagateInput( PtBlock BlockPtr, PtNodeObject NodePtr )
{
  PtPTrans	pt;
  PtNodeObject	unp,inp;
  PtVariable	var;
  PtNodeInfo	inf;

  if( NodePtr->InputNum == 0 )
	return;
  pt = (BlockPtr->PTList).Search( NodePtr->InputNum,
				  PTRANS_INPUT );
  if( pt == NULL )
  {
    wsprintf( work,"Input %u of block %s is listed on the connection list",
		    NodePtr->InputNum,(LPSTR)(BlockPtr->BlockName) );
    FatalError( work );
  }
  if( ( unp = NodePtr->UpNode ) == NULL )
  {
    wsprintf( work,"Input %u of block %s is not driven",
		   NodePtr->InputNum,(LPSTR)(BlockPtr->BlockName) );
    FatalError( work );
  }
  if( ( var = unp->NodeVariable ) == NULL )
  {
    wsprintf( work,"Variable not allocated at upnode of input %u block %s",
		   NodePtr->InputNum,(LPSTR)(BlockPtr->BlockName) );
    FatalError( work );
  }
  inf = (BlockPtr->NodeList).Search( pt->PNodeNum,OPTYPE_ALL );
  if( inf == NULL )
  {
    wsprintf( work,"Internal node %u referenced by block input %u in block %s does not exist",
		   pt->PNodeNum,NodePtr->InputNum,(LPSTR)(BlockPtr->BlockName) );
    FatalError( work );
  }
  inp = (inf->GraphPtr)->Search( pt->PIONum, NOT_VALID );
  if( inp == NULL )
  {
    wsprintf( work,"Node input %u of node %u in block %s does not exist",
		   pt->PIONum,pt->PNodeNum,(LPSTR)(BlockPtr->BlockName) );
    FatalError( work );
  }
  NodePtr->Ready = TRUE;
  inp->NodeVariable = var;
  inp->Ready = TRUE;
}

// Propagate input variables to the embedded block's
// operators. Gets a pointer to the block descriptor
void PropagateInputs( PtNodeInfo ip )
{
  PtBlock	bp;
  PtNodeObject	np;

  bp = ActBlock->SearchBlockName( ip->Operator );
  np = (ip->GraphPtr)->ForwNode;
  while( np != NULL )
  {
    if( np->InputNum != NOT_VALID )
	PropagateInput( bp,np );
    np = np->ForwNode;
  }
  np = ip->GraphPtr;
  while( np != NULL )
  {
    if( np->InputNum != NOT_VALID )
	PropagateInput( bp,np );
    np = np->BackNode;
  }
}

// Propagate one input variable from the output of the
// embedded operator to the output node
void PropagateOutput( PtBlock BlockPtr, PtNodeObject NodePtr )
{
  PtPTrans	pt;
  PtNodeObject	inp;
  PtVariable	var;
  PtNodeInfo	inf;

  pt = (BlockPtr->PTList).Search( NodePtr->OutputNum,
				  PTRANS_OUTPUT );
  if( pt == NULL )
  {
    wsprintf( work,"Output %u of block %s is listed on the connection list",
		    NodePtr->OutputNum,(LPSTR)(BlockPtr->BlockName) );
    FatalError( work );
  }
  inf = (BlockPtr->NodeList).Search( pt->PNodeNum,OPTYPE_ALL );
  if( inf == NULL )
  {
    wsprintf( work,"Internal node %u referenced by block output %u in block %s does not exist",
		   pt->PNodeNum,NodePtr->OutputNum,(LPSTR)(BlockPtr->BlockName) );
    FatalError( work );
  }
  inp = (inf->GraphPtr)->Search( NOT_VALID, pt->PIONum );
  if( inp == NULL )
  {
    wsprintf( work,"Node output %u of node %u in block %s does not exist",
		   pt->PIONum,pt->PNodeNum,(LPSTR)(BlockPtr->BlockName) );
    FatalError( work );
  }
  if( ( var = inp->NodeVariable ) == NULL )
  {
    wsprintf( work,"Variable not allocated at node %u, output %u block %s",
		   NodePtr->NodeNum,NodePtr->OutputNum,
		   (LPSTR)(BlockPtr->BlockName) );
    FatalError( work );
  }
  NodePtr->NodeVariable = var;
  NodePtr->Ready = TRUE;
}

// Propagate output variables of the embedded block to
// the outputs of the block. Gets a pointer to the block
// descriptor
void PropagateOutputs( PtNodeInfo ip )
{
  PtBlock	bp;
  PtNodeObject	np;

  bp = ActBlock->SearchBlockName( ip->Operator );
  np = (ip->GraphPtr)->ForwNode;
  while( np != NULL )
  {
    if( np->OutputNum != NOT_VALID )
	PropagateOutput( bp,np );
    np = np->ForwNode;
  }
  np = ip->GraphPtr;
  while( np != NULL )
  {
    if( np->OutputNum != NOT_VALID )
	PropagateOutput( bp,np );
    np = np->BackNode;
  }
}

// Block scheduler. Called at root level if embedded blocks
// are found. This is practically a list scheduler which
// guarantees only the precedence constraints.
void ScheduleBlocks()
{
  int		ecode;
  PtBlock	sbp,bp;
  PtNodeInfo	p,Candidate;
  PtNodeObject	nptr,np;
  PtProcessor	sp;
  PtActivity	ap,nact;
  PtListObject  SListHead,SSubList;

// First we create the ready list consisting of input nodes
// and filtering out any other operators which are not
// probes or blocks at root level.
  p = (PtNodeInfo)(ActBlock->NodeList).SubList;
  while( p != NULL )
  {
// If probe
    if( ( p->Type == OPTYPE_PROBE ) && ( p->GraphPtr != NULL ) )
    {
      nptr = (PtNodeObject)(((PtNodeObject)p->GraphPtr)->SubList);
      if( ( nptr != NULL ) && ( p->Outputs == 1 ) )
		InsertReadyList( p );
    }
    else
    if( p->Type != OPTYPE_BLOCK )
    {
      sprintf( work,"Node %u is not a block or probe",p->NodeNum );
      Error2Message( work );
    }
    p = (PtNodeInfo)p->BackLink;
  }
  if( ErrCtr != 0 )
	FatalError( "Errors detected during rootschedule" );

// Now we launch the list scheduler
// While there are unscheduled nodes
  UpScale = 0;
  DownScale = 0;
  UpdateIPF();
  while( ReadyList.SubList != NULL )
  {
    Candidate = (PtNodeInfo)ReadyList.SubList;
    ReadyList.Remove( Candidate );
    nptr = Candidate->GraphPtr;
// Forcing probes to their initial processor
    if( Candidate->Type == OPTYPE_PROBE )
    {
      sp = (PtProcessor)Candidate->SchedProc;
      ecode = BeginExecuteNode( sp,Candidate );
      if( ecode ) FatalError( "Rootschedule.BeginExecuteNode" );
      FinishExecuteNode( sp,Candidate );
    }
    else
    if( Candidate->Type == OPTYPE_BLOCK )
    {
      PropagateInputs( Candidate );
      sbp = ActBlock;
      bp = ActBlock->SearchBlockName( Candidate->Operator );
      if( bp == NULL )
      {
	wsprintf( work1,"ScheduleBlocks : Block %s not found",
			 (LPSTR)Candidate->Operator );
	FatalError( work1 );
      }
      if( !bp->Scheduled )	// Let's recurse into this block
      {
	ActBlock = bp;
// Save the NodeInfo of the block to be scheduled.
// Used by InsertVariable.
	ActBlockNodeInfo = Candidate->Alias;
	SListHead = ReadyList.ListHead;
	ReadyList.ListHead = NULL;
	SSubList = ReadyList.SubList;
	ReadyList.SubList = NULL;
	if( ecode = Schedule() )
	{
	  sprintf( work2,"Schedule error %u",ecode );
	  FatalError( work2 );
	}
	ScheduleCheck();
	ActBlock->Scheduled = TRUE;
	ActBlock->ExecTime = (ActBlock->ActivityList).MaxTime();
	Candidate->Alias->ExecTime = ActBlock->ExecTime;
	ReadyList.ListHead = SListHead;
	ReadyList.SubList = SSubList;
      }
      ActBlock = sbp;
      PropagateOutputs( Candidate );
// Check to see if the block has control input. If it has,
// we kill the variable which drives that input (as the
// control input has no internal node, nothing consumes
// the associated variable so it must be done separately)
      if( ( np = (Candidate->GraphPtr)->Search( 0,NOT_VALID ) )
	  != NULL )
		KillInputVar( np );
// Scheduling the block invocation on each processor
      sp = (PtProcessor)ProcessorList.SubList;
      while( sp != NULL )
      {
	ap = (PtActivity)(ActBlock->ActivityList).Search( sp->NodeNum );
	nact = new Activity( Candidate->NodeNum,ACT_BLOCK,
			     ap->STime, 0 );
	nact->ExecTime = Candidate->Alias->ExecTime;
	ap->Insert( nact );
	sp = (PtProcessor)sp->BackLink;
      }
      MakeSubchainsReady( nptr );	// Add descendants to readylist
      delete Candidate;	// It is done by FinishExecuteNode else
    }
  }
}

// Schedules the block tree. Starts with ROOT block,
// checks whether there are embedded blocks in the
// actual block, if not, calls schedule and schedulecheck,
// if there is, calls itself recursively to schedule
// the embedded block
void ScheduleBlockTree()
{
  int		srval;

// No blocks - can be scheduled in one step
  if( !IsDynamic )
  {
    if( srval = Schedule() )
    {
      sprintf( work2,"Schedule error %u",srval );
      FatalError( work2 );
    }
  }
  else
    ScheduleBlocks();
}

