/********* rcomp.cpp
		Rafael compiler
		Version 1.2
	    Written by Gabor Paller
				*********/

#include	<windows.h>
#include	<stdio.h>
#include	<string.h>
#include	"xserver.h"
#include	"rci.h"
#include	"rcomp.h"


/* Dialog template structure - missing from WINDOWS.H */
typedef struct {
	long	dtStyle;
	BYTE	dtItemCount;
	int	dtX;
	int	dtY;
	int	dtCX;
	int	dtCY;
	char	dtMenuName[];
	char	dtClassName[];
	char	dtCaptionText[];
} DLGTEMPLATE;
typedef DLGTEMPLATE FAR		*LPDLGTEMPLATE;

// Application instance handle
HANDLE	hInst;
// Handle to the framework's window
HWND	MainWindow;
// Handle to our dialog's window
HWND	MainDialog;
// Communication window
HWND	CommWindow;
// Status variable
int	Status;
// Shows if Ok pressed at the end
int	OkPressed;
// TRUE,if not called from the environment
BOOL	StandAlone;

// Work variable
char	work[1024];

// Source file name
char	SrcFileName[129];
// Handle of the source file
FILE	*SrcFile;

void rcomp();	// Compiler routine in rc1.cpp

// Message output file
FILE	*MsgFile;

// Error counter
unsigned ErrCtr;
// Warning counter
unsigned WarnCtr;

// Line counter
unsigned long	LineCtr;

// Fatal exit catchbuf
CATCHBUF CatchBuf;
CATCHBUF CatchBuf2;

// Incomplete ID string
char Incompl = 0xFF;

// TRUE, if XL_REQ arrived
BOOL XLREQArrived;
// TRUE, if XL_INCOMPL arrived
BOOL XLINCOMPLArrived;
// Quit arrived
BOOL QuitArrived;

// Breaks the compilation
void BreakCompile()
{
  Throw( (int FAR *)CatchBuf,1 );
}

// Exits with fatal error
void FatalError()
{
  Throw( (int FAR *)CatchBuf2,1 );
}

// Sends a message line to the output file and to the
// framework
void SendMessageLine( char *Message )
{
  RCISendMessageLine( (LPSTR)Message );
  fputs( Message,MsgFile );
  fputs( "\n",MsgFile );
}

// Updates a given text field with the given string
// Gets the ID of the text field and the text to be added
void UpdateDlgText( int DlgItem, char *text )
{
  char	*p;
  int	n;
  HWND	ctrwnd;
  char	wrk1[120];

  if( ( ctrwnd = GetDlgItem( MainDialog,DlgItem ) ) == NULL )
		return;
  n = GetWindowText( ctrwnd,wrk1,79 );

  p = wrk1;
  while( n && ( *p != ':' ) )
		++p,--n;
  if( n )
	++p,*(p++)=' ';
  strcpy( p , text );
  SetWindowText( ctrwnd,wrk1 );
}

// Gets Xlisp to execute a command. Gets the command
// string (CR-LF terminated), waits for the server
// and returns the reply. Returns NULL if the message
// loop was broken or a pointer to a 0xFF byte if
// the command was incomplete */
LPSTR ExecuteCommand( HWND Window, LPSTR Request )
{
  MSG	msg;

  XDSendRequest( Window, Request );
  XLREQArrived = FALSE;
  XLINCOMPLArrived = FALSE;
  while( !XLREQArrived && !XLINCOMPLArrived &&
	 GetMessage( &msg, NULL, 0 ,0 ) )
  {
    TranslateMessage( &msg );
    DispatchMessage( &msg );
  }

  if( XLREQArrived )
	return XDGetReply( Window );
  else
  if( XLINCOMPLArrived )
	return (LPSTR)&Incompl;
  return NULL;
}

// Debug version of the ExecuteCommand
LPSTR DRCIExecuteCommand( HWND Window, LPSTR string )
{
  LPSTR	retval;

  retval = ExecuteCommand( Window,string );
  if( ( retval == NULL ) || OkPressed )	// Message loop broken
  {
    if( ( retval != NULL ) &&
	( *(unsigned char far *)retval != 0xFF ) )
	XDDeleteReply( Window );
    BreakCompile();
  }
  return retval;
}

// Extracts the file name from the complex drive/path/file
// name
char *ExtractFileName( char *fname )
{
  char	*p;

  p = fname + _fstrlen( fname ) - 1;
  while( ( p > fname ) && ( *p != '\\' ) && ( *p != ':' ) )
		--p;
  if( p > fname )
		++p;
  return p;
}

// Processes the command line. Gets a pointer to the command
// line and a flag which shows the processing pass. The
// switches (beginning with '-') are processed. Returns
// TRU,if no source file name was found
BOOL ProcessCommandLine( LPSTR CmdLine )
{
  BOOL		SrcNameFound = FALSE;
  int		n;
  char		*p1;
  char		s[129],string[129];

  while( *CmdLine )
  {
    while( *CmdLine == ' ' )
		++CmdLine;
    p1 = string;
    n = 0;
    while( *CmdLine && ( *CmdLine != ' ' ) && ( ++n < 128) )
		*( p1++ ) = *( CmdLine++ );
    *p1 = 0;
    if( *string )
    {
      if( *string != '-' )
      {
	strncpy( SrcFileName,string,128 );
	SrcNameFound = TRUE;
      }
      else
      switch( string[1] )	/* else process switches */
      {
	  case 's':
		StandAlone = TRUE;
		break;

	  default:
		wsprintf( s , "Invalid switch : %s" , (LPSTR)string );
		MessageBox( NULL , s , "Rafael" ,
			 MB_ICONEXCLAMATION | MB_OK );
		break;
	}		/* switch */
    }			/* if( *string ) */
  }			/* while */
  return !SrcNameFound;
}

// Dialog box func for the main dialog
BOOL FAR PASCAL _export MDFunc( HWND hDlg,
		 UINT message,WPARAM wParam, LPARAM lParam )
{
  switch( message )
  {
    case WM_CLOSE:
	PostQuitMessage(0);
	return TRUE;

    case WM_INITDIALOG:
	return TRUE;

    case WM_COMMAND:
	switch( wParam )	// Child control parm
	{
	  case MD_BUTTON:
		if( HIWORD( lParam ) == BN_CLICKED )
			OkPressed = 1;
	}
	return TRUE;
  }
  return FALSE;
}

char CommClass[] = "RCOMP.COM";

// WndProc of the communication window
long FAR PASCAL _export CommWndProc( HWND hwnd, UINT message,
				     WPARAM wParam, LPARAM lParam )
{
  switch( message )
  {
    case XL_REQ:
	XLREQArrived = TRUE;
	return 0L;

    case XL_INCOMPL:
	XLINCOMPLArrived = TRUE;
	return 0L;

  }
  return DefWindowProc( hwnd, message, wParam, lParam );
}

int PASCAL WinMain( HANDLE hInstance,
		 HANDLE hPrevInstance,LPSTR lpszCmdLine,
		 int nCmdShow )
{
  RECT			rectangle;
  int			cheight,cwidth;
  int			dheight,dwidth;
  int			px,py;
  FARPROC		lpfnDlgProc;
  LPDLGTEMPLATE		lpDlgTempl;
  HANDLE		hDialog;
  long			lpDlgBaseUnits;
  WORD			wXDlg,wYDlg;
  MSG			msg;
  WNDCLASS		wndclass;

  unlink( "RCOMP.SCD" );	// Deletes the previous schedule
  OkPressed = FALSE;
  hInst = hInstance;
  StandAlone = FALSE;	// Running standalone
  if( ProcessCommandLine( lpszCmdLine ) )
  {
    MessageBox( NULL , "No source file name was given",
		    "Rafael" , MB_OK | MB_ICONSTOP );
    return 0;
  }

  lpfnDlgProc = MakeProcInstance( (FARPROC)MDFunc,hInst );
  if( RCIGetMainWindow() != NULL )
	StandAlone = FALSE;

  if( Catch( (int FAR *)CatchBuf ) )
  {
    SendMessageLine( "User break" );
    UpdateDlgText( MD_STATUSTEXT,"Break" );
  }
  else
  if( Catch( (int FAR *)CatchBuf2 ) )
  {
    UpdateDlgText( MD_STATUSTEXT,"Fatal error" );
  }
  else
  {
    if( StandAlone )
    {
      MainWindow = NULL;
      MainDialog = CreateDialog( hInst,"MDB",MainWindow,
			     lpfnDlgProc );
      if( MainDialog == NULL )
      {
	MessageBox( NULL,"Could not open compile window",
		  "Rafael" , MB_OK | MB_ICONSTOP );
	return 0;
      }
      MainWindow = MainDialog;
      RCISetMainWindow( MainWindow );
      RCISetDlgWindow( MainWindow );
    }
    else
    {
      MainWindow = RCIGetMainWindow();
      GetWindowRect( MainWindow, &rectangle );
      cheight = rectangle.bottom - rectangle.top;
      cwidth = rectangle.right - rectangle.left;
      hDialog = LoadResource( hInst,FindResource( hInst, "MDB" ,
			  RT_DIALOG ));
      lpDlgTempl = (LPDLGTEMPLATE)LockResource( hDialog );
      lpDlgBaseUnits = GetDialogBaseUnits();
      wXDlg = LOWORD( lpDlgBaseUnits );
      wYDlg = HIWORD( lpDlgBaseUnits );
      dwidth = ( lpDlgTempl->dtCX * wXDlg )/4;
      dheight = ( lpDlgTempl->dtCY * wYDlg )/8;
      px = rectangle.left + ( cwidth - dwidth )/2;
      py = rectangle.top + ( cheight - dheight )/2;
      lpDlgTempl->dtX = ( 4 * px )/wXDlg;
      lpDlgTempl->dtY = ( 8 * py )/wYDlg;
      MainDialog = CreateDialogIndirect( hInst,
	 (LPSTR)lpDlgTempl,NULL,lpfnDlgProc );
      if( MainDialog == NULL )
      {
	MessageBox( NULL,"Could not open compile window",
		  "Rafael" , MB_OK | MB_ICONSTOP );
	return 0;
      }
      MainWindow = MainDialog;
      UnlockResource( hDialog );
      RCISetDlgWindow( MainDialog );
    }		// if( StandAlone ) else ...

// Creating communication window
    wndclass.style = 0;
    wndclass.lpfnWndProc = CommWndProc;
    wndclass.cbClsExtra = 0;
    wndclass.cbWndExtra = 0;
    wndclass.hInstance = hInst;
    wndclass.hIcon = NULL;
    wndclass.hCursor = NULL;
    wndclass.hbrBackground = NULL;
    wndclass.lpszMenuName = NULL;
    wndclass.lpszClassName = CommClass;
    RegisterClass( &wndclass );

    CommWindow = CreateWindow( CommClass,NULL,WS_CHILD,
			       0,0,0,0,MainWindow,NULL,
			       hInst,NULL );

    if( XDStartServer( CommWindow ) )
    {
      MessageBox( MainDialog,"Could not start XLisp server",
		  "Rafael" , MB_OK | MB_ICONSTOP );
      DestroyWindow( CommWindow );
      DestroyWindow( MainWindow );
      return 0;
    }

// Xserver library brings the window it communicates with
// to the top after it launched the xlisp server. Unfortu-
// nately the com window is not our display window so it
// must be uncovered by another BringWindowToTop call
    BringWindowToTop( MainWindow );

    if( StandAlone )
    {
      DRCIExecuteCommand( CommWindow,"(load 'rafcomp.lsp)\n" );
      XDDeleteReply( CommWindow );
      DRCIExecuteCommand( CommWindow,"(load 'rafdb.lsp)\n" );
      XDDeleteReply( CommWindow );
      DRCIExecuteCommand( CommWindow,"(load 'rafhw.lsp)\n" );
      XDDeleteReply( CommWindow );
    }

    wsprintf( work,"Compiling %s",
	(LPSTR)ExtractFileName( SrcFileName ) );
    SetWindowText( MainDialog,(LPSTR)work );
    UpdateDlgText( MD_COMPTEXT,ExtractFileName( SrcFileName ) );
    UpdateDlgText( MD_ERRTEXT,"0" );
    UpdateDlgText( MD_WARNTEXT,"0" );
    Status = 0;

    OkPressed = 0;
    if( ( SrcFile = fopen( SrcFileName,"r" ) ) == NULL )
    {
      wsprintf( work,"No file : %s",
	      (LPSTR)ExtractFileName( SrcFileName ) );
      SetWindowText( GetDlgItem( MainDialog, MD_COMPTEXT ),
		   work );
    }
    else
    if( ( MsgFile = fopen( "rcomp.msg","w" ) ) == NULL )
    {
      strcpy( work,"Cannot open message file" );
      SetWindowText( GetDlgItem( MainDialog, MD_COMPTEXT ),
		   work );
    }
    else
	rcomp();
    if( ErrCtr == 0 )
    {
      sprintf( work,"No errors compiling %s",SrcFileName );
      SendMessageLine( work );
    }
    UpdateDlgText( MD_STATUSTEXT,"End run" );
  }
  fclose( MsgFile );
  SetWindowText( GetDlgItem( MainDialog,MD_BUTTON ), "Ok" );
  OkPressed = FALSE;
  if( !Status )		// If not terminated with error
			// then wait for the Ok keypress
    while( !OkPressed && GetMessage( &msg, NULL, 0, 0 ) )
    {
      if( !IsDialogMessage( MainDialog, &msg ) )
      {
	TranslateMessage( &msg );
	DispatchMessage( &msg );
      }
    }
  RCISetDlgWindow( NULL );
  DestroyWindow( CommWindow );
  DestroyWindow( MainDialog );
  XDTerminateServer();
  if( !StandAlone )
	FreeResource( hDialog );

  FreeProcInstance( lpfnDlgProc );
  RCISendEndCompile();
  return Status;
}
