/******* robject.cpp
		 Rafael object library
		Written by Gabor Paller
		    Part of RCOMP
					**********/

#include	<windows.h>
#include	<stdio.h>
#include	<string.h>
#include	<alloc.h>
#include	"robject.h"

// ListObject. Implements a list of another lists. Basic node
// element

// Constructor. Gets the node number
ListObject::ListObject( NODETYPE node )
{
  NodeNum = node;
  BackLink = NULL;
  ForwLink = NULL;
  SubList = NULL;
  ListHead = NULL;
}

// Inserts one element into the sublist. Gets the node to
// be inserted
void ListObject::Insert( PtListObject NewObject )
{
  if( SubList != NULL )
	SubList->ForwLink = NewObject;
  NewObject->BackLink = SubList;
  NewObject->ForwLink = NULL;
  SubList = NewObject;
  if( ListHead == NULL )
	ListHead = NewObject;
}

// Inserts an element to a given position of the list. Gets
// a pointer to the element and the position to which the
// element will be inserted. The list elements are numbered
// from the tail of the list, the end of list is position 0.
void ListObject::InsertFromTail( PtListObject NewObject,
			unsigned Pos )
{
  PtListObject	p,p1 = NULL;

  p = SubList;
  while( ( p != NULL ) && ( ( Pos-- ) > 0 ) )
	p1 = p, p = p->BackLink;
  if( p == NULL )	// Inserting to the head of the list
  {
    NewObject->BackLink = NULL;
    NewObject->ForwLink = p1;
    if( p1 != NULL )
	p1->BackLink = NewObject;
    ListHead = NewObject;
  }
  else			// Inserting to the middle or tail
  {
    NewObject->ForwLink = p1;
    NewObject->BackLink = p;
    p->ForwLink = NewObject;
    if( p1 == NULL )		// Inserting to the tail
	SubList = NewObject;
    else
	p1->BackLink = NewObject;
  }
  if( SubList == NULL )
	SubList = NewObject;
}

// Removes an element from the sublist. Gets pointer
// to the element to be removed
void ListObject::Remove( PtListObject OldObject )
{
  if( OldObject->BackLink != NULL )
    OldObject->BackLink->ForwLink = OldObject->ForwLink;
  if( OldObject->ForwLink != NULL )
    OldObject->ForwLink->BackLink = OldObject->BackLink;
  if( SubList == OldObject )
	SubList = OldObject->BackLink;
  if( ListHead == OldObject )
		ListHead = OldObject->ForwLink;
}

// Searches a given node in the sublist. Gets the node number,
// returns a pointer to the node found or NULL.
PtListObject ListObject::Search( NODETYPE snode )
{
  PtListObject	p;

  p = SubList;
  while( p != NULL )
	if( p->NodeNum == snode )
		break;
	else
	  p = p->BackLink;
  return p;
}

// Chunk object. Used to represent the actual memory map.
// Stores the variable ID, the start address and the
// size of the variable.

// Chunk constructor. Gets the processor ID, the variable ID,
// the start address and the size
Chunk::Chunk( int proc, unsigned long id, unsigned long start,
	      unsigned long size ) : ListObject( proc )
{
  ID = id;
  StartAddr = start;
  Size =  size;
  Avail = size;
}

// Default Chunk constructor
Chunk::Chunk() : ListObject( 0 )
{
  ID = 0;
  StartAddr = 0L;
  Size =  0L;
  Avail = 0L;
}

// Allocates a given amount of memory. Returns the start
// address of the chunk allocated or CHK_NOMEM
// if no memory is available
unsigned long Chunk::Alloc( unsigned long ReqMem )
{
  unsigned long	HighLimit,LowLimit;
  PtChunk p;

  if( Avail < ReqMem )
		return CHK_NOMEM;
  LowLimit = 0L;
  p = (PtChunk)SubList;
  while( p != NULL )
  {
    HighLimit = p->StartAddr - 1;
    if( ( HighLimit - LowLimit + 1 ) >= ReqMem )
		return LowLimit;
    else
    {
      LowLimit = p->StartAddr + p->Size;
      p = (PtChunk)p->BackLink;
    }
  }
  HighLimit = Size - 1;
  if( ( HighLimit - LowLimit + 1 ) >= ReqMem )
		return LowLimit;
  return CHK_NOMEM;
}

// Inserts a chunk into the list. The chunks are ordered
// in ascending order by StartAddr
void Chunk::Insert( PtChunk Ptr )
{
  PtChunk	p;
  PtChunk	p1;

  p = (PtChunk)SubList;
  p1 = NULL;
  while( p != NULL )
	if( p->StartAddr > Ptr->StartAddr )
		break;
	else
	{
	  p1 = p;
	  p = (PtChunk)p->BackLink;
	}
  Ptr->BackLink = (PtListObject)p;
  Ptr->ForwLink = (PtListObject)p1;
  if( p1 != NULL )
	p1->BackLink = (PtListObject)Ptr;
  else
	SubList = (PtListObject)Ptr;
  if( p != NULL )
	p->ForwLink = (PtListObject)Ptr;
  Avail -= Ptr->Size;
}

// Removes a chunk from the list
void Chunk::Remove( PtChunk Ptr )
{
  ListObject::Remove( (PtListObject)Ptr );
  Avail += Ptr->Size;
}

// Variable object. Used to track the node variables.
// Stores the variable ID, the number of consuming, the
// physical address of the variable and the associated
// memory map chunk.

// Constructor. Gets the variable ID
Variable::Variable( int proc, unsigned long id ) : ListObject( proc )
{
  ID = id;
  Address = 0L;
  Consumed = 0;
  AssocChunk = NULL;
  GraphPtr = NULL;
  ProcPtr = NULL;
  BlockPtr = NULL;
  VarLink = this;
}

// Default constructor for Variable
Variable::Variable() : ListObject( 0 )
{
  ID = 0;
  Address = 0L;
  Consumed = 0;
  AssocChunk = NULL;
  GraphPtr = NULL;
  ProcPtr = NULL;
  BlockPtr = NULL;
  VarLink = this;
}

// Inserts a variable. The list is sorted by address
void Variable::Insert( PtVariable Ptr )
{
  PtVariable	p;
  PtVariable	p1;
  unsigned long	MemTop;

  MemTop = Ptr->Address + Ptr->Size;
  if( MemTop > Address )
	Address = MemTop;
  p = (PtVariable)SubList;
  p1 = NULL;
  while( p != NULL )
	if( p->Address < Ptr->Address )
		break;
	else
	{
	  p1 = p;
	  p = (PtVariable)p->BackLink;
	}
  Ptr->BackLink = (PtListObject)p;
  Ptr->ForwLink = (PtListObject)p1;
  if( p == NULL )
	ListHead = (PtListObject)Ptr;
  if( p1 != NULL )
	p1->BackLink = (PtListObject)Ptr;
  else
	SubList = (PtListObject)Ptr;
  if( p != NULL )
	p->ForwLink = (PtListObject)Ptr;
}

// Channel object. The object is used to store the
// activities on the communication channels.
// Used in two-level list where the first level list elements
// store the actual state of the channel, their sublists
// administrate the history of the activities on the
// channel.
Channel::Channel( NODETYPE ChNum ) : ListObject( ChNum )
{
  SendStartOccupied = 0L;
  SendEndOccupied = 0L;
  RecStartOccupied = 0L;
  RecEndOccupied = 0L;
}

// Default constructor
Channel::Channel() : ListObject( 0 )
{
  SendStartOccupied = 0L;
  SendEndOccupied = 0L;
  RecStartOccupied = 0L;
  RecEndOccupied = 0L;
}

void Channel::Insert( Channel *NCH )
{
  ListObject::Insert( (PtListObject)NCH );
}

PtChannel Channel::Search( NODETYPE Node )
{
  return (PtChannel)ListObject::Search( Node );
}

// ExecTimeList object. Used to store the execution times
// for the processors available.
ExecTimeList::ExecTimeList( char D[], unsigned long E ) : ListObject( 0 )
{
  _fstrncpy( Dev,D,31 );
  ExecTime = E;
}

void ExecTimeList::Insert( PtExecTimeList ETL )
{
  ListObject::Insert( (PtListObject)ETL );
}

// Searches in the sublist for an entry whose processor
// name is given.
PtExecTimeList ExecTimeList::SearchDevice( char D[] )
{
  PtExecTimeList p;

  p = (PtExecTimeList)SubList;
  while( p != NULL )
  {
    if( _fstrncmp( p->Dev,D,31 ) == 0 )
		return p;
    p = (PtExecTimeList)p->BackLink;
  }
  return p;
}

// Clears all the entries in the sublist.
void ExecTimeList::ClearSubList()
{
  PtExecTimeList p,p1;

  p = (PtExecTimeList)SubList;
  while( p != NULL )
  {
    p1 = p;
    p = (PtExecTimeList)p->BackLink;
    delete p1;
  }
  SubList = NULL;
}

// NodeObject. Implements multi-input node tree.
// The inputs are numbered, the multiple inputs are implemented
// as multiple node entries with the same node number and
// different input and output number. Each node is represented
// with at least 2 entries in a given connection sublist.
// The outputs of other nodes point to NodeObjects with
// given input numbers and output numbers set to NOT_VALID.
// The outputs of the operator are represented by NodeObjects
// whose input numbers are set to NOT_VALID and their output
// numbers denote the number of the output. Subsequent
// nodes are inserted ONLY into output nodes. The nodes
// belonging to one operator are linked into one list by
// the ForwNode and BackNode data members of NodeObject.

// Constructor. Gets the node number, the input and output
// numbers.
NodeObject::NodeObject( NODETYPE snode, INPUTTYPE INum,
			     INPUTTYPE ONum ) :
				 ListObject( snode )
{
  ForwNode = NULL;
  BackNode = NULL;
  UpNode = NULL;
  InfoNode = NULL;
  Ready = FALSE;
  Blocked = FALSE;
  Auto = TRUE;
  InputNum = INum;
  OutputNum = ONum;
  NodeVariable = NULL;
  Size = 0L;
  Aux = 0L;
  *TypeName = 0;
  VarReadyTime = 0L;
  UIPF = 0;
  DIPF = 0;
  UpScale = 0;
  DownScale = 0;
}

// Inserts a node into a sublist
void NodeObject::Insert( PtNodeObject NewNode )
{
  ListObject::Insert( (PtListObject)NewNode );
  NewNode->UpNode = this;
}

// Inserts a node into the operator chain
void NodeObject::InsertOp( PtNodeObject NewNode )
{
  PtNodeObject 	ListEnd;
  PtNodeObject	p;

  p = ForwNode;
  ListEnd = this;
  while( p != NULL )
  {
    ListEnd = p;
    p = p->ForwNode;
  }

  ListEnd->ForwNode = NewNode;
  NewNode->BackNode = ListEnd;
}


// Returns TRUE if the node is ready to execute so that
// the Ready fields of all the input nodes are TRUE
BOOL NodeObject::IsReady()
{
  PtNodeObject p;

  p = this;
  while( p != NULL )
    if( ( p->InputNum != NOT_VALID ) && !( p->Ready ) )
	return FALSE;
    else
	p = p->ForwNode;

  p = this;
  while( p != NULL )
    if( ( p->InputNum != NOT_VALID ) && !( p->Ready ) )
	return FALSE;
    else
	p = p->BackNode;
  return TRUE;
}

// Resets the flag variables of all the nodes in the
// operator chain
void NodeObject::ResetFlags()
{
  PtNodeObject p;

  p = this;
  while( p != NULL )
  {
    p->Ready = FALSE;
    p->Blocked = FALSE;
    p = p->ForwNode;
  }

  p = this;
  while( p != NULL )
  {
    p->Ready = FALSE;
    p->Blocked = FALSE;
    p = p->BackNode;
  }
}

// Searches for a given input and output node in the object's
// own list. Returns pointer to the object or NULL if not found.
PtNodeObject NodeObject::Search( INPUTTYPE Inp, INPUTTYPE Out )
{
  PtNodeObject p;

  p = this;
  while( p != NULL )
    if( ( p->InputNum == Inp ) && ( p->OutputNum == Out ) )
	return p;
    else
	p = p->ForwNode;
  p = this;
  while( p != NULL )
    if( ( p->InputNum == Inp ) && ( p->OutputNum == Out ) )
	return p;
    else
	p = p->BackNode;
  return p;
}

// Substitutes all the 'oldname' type names with 'newname'
// type name
void NodeObject::SubstituteTypeName( LPSTR oldname, LPSTR newname )
{
  PtNodeObject p;

  p = this;
  while( p != NULL )
  {
    if( _fstrcmp( p->TypeName, oldname ) == 0 )
	_fstrcpy( p->TypeName, newname );
    p = p->BackNode;
  }
  p = ForwNode;
  while( p != NULL )
  {
    if( _fstrcmp( p->TypeName, oldname ) == 0 )
	_fstrcpy( p->TypeName, newname );
    p = p->ForwNode;
  }
}

// DelayInputObject. DelayInputObject is a NodeObject which has a
// modified IsReady method for supporting delay inputs
void DelayInputObject::DelayInputObject( NODETYPE snode,
			      INPUTTYPE INum,INPUTTYPE ONum ) :
		   NodeObject( snode,INum,ONum )
{
}

// Checks whether the delay input is ready to execute.
// As delay outputs are placed into the ready list at
// the beginning, this routine is called only in case
// of delay inputs. Delay inputs are ready if their
// delay output parts have already been executed.
BOOL DelayInputObject::IsReady()
{
  PtNodeObject p;

  p = this;
  while( p != NULL )
    if( !( p->Ready ) )
	return FALSE;
    else
	p = p->ForwNode;

  p = this;
  while( p != NULL )
    if( !( p->Ready ) )
	return FALSE;
    else
	p = p->BackNode;
  return TRUE;
}

// DelayOutputObject. DelayOutputObject is a NodeObject which
// has a modified IsReady method for supporting delay outputs
void DelayOutputObject::DelayOutputObject( NODETYPE snode,
			      INPUTTYPE INum,INPUTTYPE ONum ) :
		   NodeObject( snode,INum,ONum )
{
}

// Checks whether the delay output is ready to execute.
// Delay outputs are always ready to execute
BOOL DelayOutputObject::IsReady()
{
  return Ready;
}

// NodeInfo object. This object is used in a linked
// list which acts as a dictionary of nodes.
// NodeInfo stores the number of inputs and outputs
// of the node, the name of the operator associated
// with it and a pointer to a string which stores
// the parameter list

// Constructor. Gets the node number,the node type,
// the number of the inputs and outputs,the name of
// the operator
NodeInfo::NodeInfo( NODETYPE snode ,int OpType,
		    int IN ,int ON , LPSTR opname) :
			ListObject( snode )
{
  Inputs = IN;
  Outputs = ON;
  Type = OpType;
  GraphPtr = NULL;
  _fstrcpy( (LPSTR)Operator,opname );
  Executed = FALSE;
  Visited = FALSE;
  ASAP = 0L;
  ALAP = 0L;
  Aux = 0L;
  StartTime = 0L;
  FinishTime = 0L;
  Alias = NULL;
  SchedProc = NULL;
}

// Default constructor
NodeInfo::NodeInfo() : ListObject(0)
{
  Inputs = 0;
  Outputs = 0;
  Type = 0;
  GraphPtr = NULL;
  *Operator = 0;
  Executed = FALSE;
  Visited = FALSE;
  ASAP = 0L;
  ALAP = 0L;
  Aux = 0L;
  StartTime = 0L;
  FinishTime = 0L;
  Alias = NULL;
  SchedProc = NULL;
}

// Inserts an element. Gets a pointer to the object
void NodeInfo::Insert( PtNodeInfo Ptr )
{
  ListObject::Insert( (PtListObject)Ptr );
}

// Searches for an element. Gets the node number and the type
// mask
NodeInfo *NodeInfo::Search( NODETYPE NodeNum , int Mask )
{
  PtNodeInfo	p;

  p = (PtNodeInfo)SubList;
  while( p != NULL )
    if( ( p->NodeNum == NodeNum ) && ( p->Type & Mask ) )
	break;
    else
	p = (PtNodeInfo)p->BackLink;
  return p;
}

// Calculates the possible start time of the operator
// Steps through the input connections and writes
// the maximum VarReadyTime into the StartTime member
// variable
void NodeInfo::CalculateStartTime()
{
  PtNodeObject	np;
  unsigned long STime = 0L;

  np = GraphPtr;
  while( np != NULL )
  {
    if( ( np->InputNum != NOT_VALID ) &&
	( np->VarReadyTime > STime ) )
		STime = np->VarReadyTime;
    np = np->ForwNode;
  }
  np = GraphPtr;
  while( np != NULL )
  {
    if( ( np->InputNum != NOT_VALID ) &&
	( np->VarReadyTime > STime ) )
		STime = np->VarReadyTime;
    np = np->BackNode;
  }
  StartTime = STime;
}

// Gets the upscale and downscale factors of the
// output nodes. Gets two pointers to the upscale
// and downscale factors and updates them appropriately
// Returns FALSE if the UpScale and DownScale fields
// of the node has not been filled appropriately
BOOL NodeInfo::GetScales( unsigned *pus,unsigned *pds )
{
  PtNodeObject	np;
  float	fac,maxfac;

  maxfac = -1;
  np = GraphPtr;
  while( np != NULL )
  {
    fac = (float)np->UpScale/(float)np->DownScale;
    if( fac > maxfac )
    {
      maxfac = fac;
      *pus = np->UpScale;
      *pds = np->DownScale;
    }
    np = np->ForwNode;
  }
  np = GraphPtr;
  while( np != NULL )
  {
    fac = (float)np->UpScale/(float)np->DownScale;
    if( np->InputNum == NOT_VALID )
    {
      maxfac = fac;
      *pus = np->UpScale;
      *pds = np->DownScale;
    }
    np = np->BackNode;
  }
  return TRUE;
}

// Activity object. This object is used to store
// the schedule sequence for each processor.
// Used in two-level list where the first list
// stores the processor IDs and the activities
// are inserted into their subview list

// Activity constructor. Gets the node number, the
// activity code, the scheduling time and the aux
// variable.
Activity::Activity( NODETYPE snode, int ACTID,
		    unsigned long ST, int AUX ) :
		    ListObject( snode )
{
  ActId = ACTID;
  STime = ST;
  CommInProgress = 0L;
  ExecTime = 0L;
  VarField = NULL;
  Aux = AUX;
  PartnerAct = NULL;
  UpScale = 1;
  DownScale = 1;
}

// Default constructor
Activity::Activity() : ListObject( 0 )
{
  ActId = 0;
  STime = 0L;
  CommInProgress = 0L;
  ExecTime = 0L;
  VarField = NULL;
  Aux = 0;
  PartnerAct = NULL;
  UpScale = 1;
  DownScale = 1;
}

// Inserts an activity into our sublist
void Activity::Insert( PtActivity Ptr )
{
  ListObject::Insert( (PtListObject)Ptr );
}

// Returns the maximal time field of the sublist
// activities
unsigned long Activity::MaxTime()
{
  PtActivity	ap = (PtActivity)SubList;
  unsigned long	maxtime = 0L;

  while( ap != NULL )
  {
    if( ap->STime > maxtime )
	maxtime = ap->STime;
    ap = (PtActivity)ap->BackLink;
  }
  return maxtime;
}

// Processor object. This object is used to create
// the system descriptor list. This list contains
// processor objects which denote physical devices.

// Processor object construcor. Gets :
// processor number (comm target), memory available
// (comm cost) and the device name.
Processor::Processor( NODETYPE Proc, unsigned long MEM,
		      char DEV[] ) :
		      ListObject( Proc )
{
  OccupiedTime = 0L;
  Memory = MEM;
  Aux = 0L;
  UpScale = 0;
  DownScale = 0;
  _fstrncpy( (LPSTR)Device,(LPSTR)DEV,31 );
}

// Inserts a Processor node into a sublist
void Processor::Insert( PtProcessor Ptr )
{
  ListObject::Insert( (PtListObject)Ptr );
}

// Parameter translation object
// Constructor. Gets input/output number, internal node
// number and its I/O number, I/O direction flag and type
// name
PTrans::PTrans( NODETYPE SNode, NODETYPE PNode, NODETYPE PIO,
		int D, char tname[] ) :
 ListObject( SNode )
{
  Dir = D;
  PNodeNum = PNode;
  PIONum = PIO;
  _fstrncpy( Type,tname,31 );
}

// Default constructor
PTrans::PTrans() : ListObject(0)
{
  Dir = 0;
  PNodeNum = 0;
  PIONum = 0;
  *Type = 0;
}

// Insert method
void PTrans::Insert( PtPTrans Object )
{
  ListObject::Insert( (PtListObject)Object );
}

// Searches for the given I/O number and returns a
// pointer to its descriptor. Gets I/O number and I/O flag.
PtPTrans PTrans::Search( NODETYPE SNode, int D )
{
  PtPTrans	p;

  p = (PtPTrans)SubList;
  while( p != NULL )
    if( ( p->Dir == D ) && ( p->NodeNum == SNode ) )
	break;
    else
	p = (PtPTrans)p->BackLink;
  return p;
}

// Block object. Stores all the information about
// a block.
Block::Block( char Name[],BOOL ir ) : ListObject( 0 )
{
  _fstrncpy( BlockName,Name,31 );
  ExecTime = 0L;
  InputNum = OutputNum = 0;
  Scheduled = FALSE;
  ConstrGenerated = FALSE;
  OpGenerated = FALSE;
  DestrGenerated = FALSE;
  IsRoot = ir;
}

// Insert method
void Block::Insert( PtBlock BPtr )
{
  ListObject::Insert( (PtListObject)BPtr );
}

// Searches for a block name beginning with the current
// block
PtBlock Block::SearchBlockName( char Name[] )
{
  PtBlock p = (PtBlock)SubList;

  while( p != NULL )
    if( _fstrncmp( p->BlockName,Name,31 ) == 0 )
	break;
    else
	p = (PtBlock)p->BackLink;
  return p;
}

/****************** END HEADER - ROBJECT.CPP ***************/
